"""
This module provides day-count convention definitions and functionlity.
"""

__all__ = ["DCC", "DCCRegistry"]

import calendar
import datetime
from decimal import Decimal
from typing import Callable, Dict, Iterable, List, NamedTuple, Optional, Set, Union

from dateutil.relativedelta import relativedelta

from .commons.numbers import ONE, ZERO
from .commons.zeitgeist import Date
from .currencies import Currencies, Currency
from .monetary import Money

#: Defines a type alias for day count fraction calculation functions.
DCFC = Callable[[Date, Date, Date, Optional[Decimal]], Decimal]


def _as_ccys(codes: Set[str]) -> Set[Currency]:
    """
    Converts a set of currency codes to a set of currencies.
    """
    return {Currencies[c] for c in codes}


def _get_date_range(start: Date, end: Date) -> Iterable[Date]:
    """
    Returns a generator of dates falling into range within the given period (``end`` is exclusive).

    :param start: The start date of the period.
    :param end: The end date of the period.
    :return: A generator of dates.
    """
    for i in range((end - start).days):
        yield start + datetime.timedelta(days=i)


def _get_actual_day_count(start: Date, end: Date) -> int:
    """
    Counts the actual number of days in the given period.

    :param start: The start date of the period.
    :param end: The end date of the period.
    :return: The number of days in the given period.

    >>> _get_actual_day_count(datetime.date(2017, 1, 1), datetime.date(2017, 1, 1))
    0
    >>> _get_actual_day_count(datetime.date(2017, 1, 1), datetime.date(2017, 1, 2))
    1
    """
    return (end - start).days


def _has_leap_day(start: Date, end: Date) -> bool:
    """
    Indicates if the range has any leap day.
    """
    ## Get all leap years:
    years = {year for year in range(start.year, end.year + 1) if calendar.isleap(year)}

    ## Check if any of the lap day falls in our range:
    for year in years:
        ## Construct the leap day:
        leapday = datetime.date(year, 2, 29)

        ## Is the leap date in the range?
        if start <= leapday <= end:
            ## Yes, the leap day is within the date range. Return True:
            return True

    ## No leap day in the range, return False:
    return False


def _is_last_day_of_month(date: Date) -> bool:
    """
    Indicates if the date is the last day of the month.
    """
    return date.day == calendar.monthrange(date.year, date.month)[1]


def _last_payment_date(start: Date, asof: Date, frequency: Union[int, Decimal], eom: Optional[int] = None) -> Date:
    """
    Returns the last coupon payment date.

    >>> _last_payment_date(datetime.date(2014,  1,  1), datetime.date(2015, 12, 31), 1)
    datetime.date(2015, 1, 1)

    >>> _last_payment_date(datetime.date(2015,  1,  1), datetime.date(2015, 12, 31), 1)
    datetime.date(2015, 1, 1)

    >>> _last_payment_date(datetime.date(2014,  1,  1), datetime.date(2015, 12, 31), 2)
    datetime.date(2015, 7, 1)

    >>> _last_payment_date(datetime.date(2014,  1,  1), datetime.date(2015,  8, 31), 2)
    datetime.date(2015, 7, 1)

    >>> _last_payment_date(datetime.date(2014,  1,  1), datetime.date(2015,  4, 30), 2)
    datetime.date(2015, 1, 1)

    >>> _last_payment_date(datetime.date(2014,  6,  1), datetime.date(2015,  4, 30), 1)
    datetime.date(2014, 6, 1)

    >>> _last_payment_date(datetime.date(2008,  7,  7), datetime.date(2015, 10,  6), 4)
    datetime.date(2015, 7, 7)

    >>> _last_payment_date(datetime.date(2014, 12,  9), datetime.date(2015, 12,  4), 1)
    datetime.date(2014, 12, 9)

    >>> _last_payment_date(datetime.date(2012, 12, 15), datetime.date(2016,  1,  6), 2)
    datetime.date(2015, 12, 15)

    >>> _last_payment_date(datetime.date(2012, 12, 15), datetime.date(2015, 12, 31), 2)
    datetime.date(2015, 12, 15)
    """
    ## Make sure that we have eom:
    eom = eom or start.day

    ## Get the starting month:
    s_month = start.month

    ## Get the period:
    period = int(12 / frequency)

    ## Get the current day, month and year:
    c_day, c_month, c_year = asof.day, asof.month, asof.year

    ## Get the payment schedule:
    schedule = sorted([i > 0 and i or 12 for i in sorted([(i + s_month) % 12 for i in range(0, 12, period)])])

    ## Filter out previous:
    future = [month for month in schedule if (month < c_month) or (month == c_month and eom <= c_day)]

    ## Get the previous month and year:
    p_year, p_month = (c_year, future[-1]) if future else (c_year - 1, schedule[-1])

    ## Return the date:
    if p_year < 1 or p_month < 1 or eom < 1:
        return start

    ## Construct and return the date safely:
    return _construct_date(p_year, p_month, eom)


def _next_payment_date(start: Date, frequency: Union[int, Decimal], eom: Optional[int] = None) -> Date:
    """
    Returns the last coupon payment date.

    >>> _next_payment_date(datetime.date(2014,  1,  1), 1, None)
    datetime.date(2015, 1, 1)

    >>> _next_payment_date(datetime.date(2014,  1,  1), 1, 15)
    datetime.date(2015, 1, 15)
    """
    ## Get the number of months to move forward:
    months = int(12 / frequency)

    ## Find the next date:
    nextdate = start + relativedelta(months=months)

    ## Do we have any end of month?
    if eom:
        try:
            nextdate = nextdate.replace(day=eom)
        except ValueError:
            pass

    ## Done, return:
    return nextdate


def _construct_date(year: int, month: int, day: int) -> Date:
    """
    Constructs and returns date safely.
    """
    if year <= 0 or month <= 0 or day <= 0:
        raise ValueError("year, month and day must be greater than 0.")
    try:
        return datetime.date(year, month, day)
    except ValueError as exc:
        if str(exc) == "day is out of range for month":
            return _construct_date(year, month, day - 1)
        else:
            raise exc


class DCC(NamedTuple):
    """
    Defines a day count convention model.
    """

    #: Defines the name of the day count convention.
    name: str

    #: Defines a set of alternative names of the day count convention.
    altnames: Set[str]

    #: Defines a set of currencies which are known to use this convention by default.
    currencies: Set[Currency]

    #: Defines the day count fraction calculation method function.
    calculate_fraction_method: DCFC

    def calculate_fraction(self, start: Date, asof: Date, end: Date, freq: Optional[Decimal] = None) -> Decimal:
        """
        Calculates the day count fraction based on the underlying methodology after performing some general checks.
        """
        ## Checks if dates are provided properly:
        if not start <= asof <= end:
            ## Nope, return 0:
            return ZERO

        ## Cool, we can proceed with calculation based on the methodology:
        return self[3](start, asof, end, freq)

    def calculate_daily_fraction(self, start: Date, asof: Date, end: Date, freq: Optional[Decimal] = None) -> Decimal:
        """
        Calculates daily fraction.
        """
        ## Get t-1 for asof:
        asof_minus_1 = asof - datetime.timedelta(days=1)

        ## Get the yesterday's factor:
        if asof_minus_1 < start:
            yfact = ZERO
        else:
            yfact = self.calculate_fraction_method(start, asof_minus_1, end, freq)

        ## Get today's factor:
        tfact = self.calculate_fraction_method(start, asof, end, freq)

        ## Get the factor and return:
        return tfact - yfact

    def interest(
        self,
        principal: Money,
        rate: Decimal,
        start: Date,
        asof: Date,
        end: Optional[Date] = None,
        freq: Optional[Decimal] = None,
    ) -> Money:
        """
        Calculates the accrued interest.
        """
        return principal * rate * self.calculate_fraction(start, asof, end or asof, freq)

    def coupon(
        self,
        principal: Money,
        rate: Decimal,
        start: Date,
        asof: Date,
        end: Date,
        freq: Union[int, Decimal],
        eom: Optional[int] = None,
    ) -> Money:
        """
        Calculates the accrued interest for the coupon payment.

        This method is primarily used for bond coupon accruals which assumes the start date to be the first of regular
        payment schedules.
        """
        ## Find the previous and next payment dates:
        prevdate = _last_payment_date(start, asof, freq, eom)
        nextdate = _next_payment_date(prevdate, freq, eom)

        ## Calculate the interest and return:
        return self.interest(principal, rate, prevdate, asof, nextdate, Decimal(freq))


class DCCRegistryMachinery:
    """
    Provides the day count registry model.

    >>> principal = Money.of(Currencies["USD"], Decimal(1000000), datetime.date.today())
    >>> start = datetime.date(2007, 12, 28)
    >>> end = datetime.date(2008, 2, 28)
    >>> rate = Decimal(0.01)
    >>> dcc = DCCRegistry.find("Act/Act")
    >>> round(dcc.calculate_fraction(start, end, end), 14)
    Decimal('0.16942884946478')
    >>> dcc.interest(principal, rate, start, end, end).qty
    Decimal('1694.29')
    >>> dcc.interest(principal, rate, end, start, start).qty
    Decimal('0.00')
    """

    def __init__(self) -> None:
        """
        Initializes the registry.
        """
        ## Define the main registry buffer:
        self._buffer_main: Dict[str, DCC] = {}

        ## Defines the registry buffer for alternative DCC names:
        self._buffer_altn: Dict[str, DCC] = {}

    def _is_registered(self, name: str) -> bool:
        """
        Checks if the given name is ever registered before.
        """
        return name in self._buffer_main or name in self._buffer_altn

    def register(self, dcc: DCC) -> None:
        """
        Attempts to register the given day count convention.
        """
        ## Check if the main name is ever registered before:
        if self._is_registered(dcc.name):
            ## Yep, raise a TypeError:
            raise TypeError(f"Day count convention '{dcc.name}' is already registered")

        ## Add to the main buffer:
        self._buffer_main[dcc.name] = dcc

        ## Check if there is any registry conflict:
        for name in dcc.altnames:
            ## Check if the name is ever registered:
            if self._is_registered(name):
                ## Yep, raise a TypeError:
                raise TypeError(f"Day count convention '{dcc.name}' is already registered")

            ## Register to the alternative buffer:
            self._buffer_altn[name] = dcc

    def _find_strict(self, name: str) -> Optional[DCC]:
        """
        Attempts to find the day count convention by the given name.
        """
        return self._buffer_main.get(name) or self._buffer_altn.get(name)

    def find(self, name: str) -> Optional[DCC]:
        """
        Attempts to find the day count convention by the given name.

        Note that all day count conventions are registered under stripped, uppercased names. Therefore,
        the implementation will first attempt to find by given name as is. If it can not find it, it will
        strip and uppercase the name and try to find it as such as a last resort.
        """
        return self._find_strict(name) or self._find_strict(name.strip().upper())

    @property
    def registry(self) -> List[DCC]:
        """
        Returns the main registry values.
        """
        return list(self._buffer_main.values())

    @property
    def table(self) -> Dict[str, DCC]:
        """
        Returns a lookup table for available day count conventions.
        """
        return {**{k: v for k, v in self._buffer_main.items()}, **{k: v for k, v in self._buffer_altn.items()}}


#: Defines the default DCC registry.
DCCRegistry = DCCRegistryMachinery()


def dcc(name: str, altnames: Optional[Set[str]] = None, ccys: Optional[Set[Currency]] = None) -> Callable[[DCFC], DCFC]:
    """
    Registers a day count fraction calculator under the given names and alternative names (if any).

    :param name: The name of the day count convention.
    :param altnames: A set of alternative names of the day count convention, if any.
    :param ccys: A set of currencies which are known to use this convention by default, if any.
    :return: Registered day count fraction calculation function.
    """

    def register_and_return_dcfc(func: DCFC) -> DCFC:
        """
        Registers the given day count fraction calculator and returns it.

        :param func: Day count fraction calculation function to be registered.
        :return: Registered day count fraction calculation function.
        """
        ## Create the DCC instance:
        dcc = DCC(name, altnames or set([]), ccys or set([]), func)

        ## Attempt to register the DCC:
        DCCRegistry.register(dcc)

        ## Attach the dcc instance to the day count fraction calculation function (for whatever it is worth):
        setattr(func, "__dcc", dcc)

        ## Done, return the function (if above statment did not raise any exceptions):
        return func

    return register_and_return_dcfc


@dcc("Act/Act", {"Actual/Actual", "Actual/Actual (ISDA)"})
def dcfc_act_act(start: Date, asof: Date, end: Date, freq: Optional[Decimal] = None) -> Decimal:
    """
    Computes the day count fraction for "Act/Act" convention.

    :param start: The start date of the period.
    :param asof: The date which the day count fraction to be calculated as of.
    :param end: The end date of the period (a.k.a. termination date).
    :param freq: The frequency of payments in a year.
    :return: Day count fraction.

    >>> ex1_start, ex1_asof = datetime.date(2007, 12, 28), datetime.date(2008, 2, 28)
    >>> ex2_start, ex2_asof = datetime.date(2007, 12, 28), datetime.date(2008, 2, 29)
    >>> ex3_start, ex3_asof = datetime.date(2007, 10, 31), datetime.date(2008, 11, 30)
    >>> ex4_start, ex4_asof = datetime.date(2008, 2, 1), datetime.date(2009, 5, 31)
    >>> round(dcfc_act_act(start=ex1_start, asof=ex1_asof, end=ex1_asof), 14)
    Decimal('0.16942884946478')
    >>> round(dcfc_act_act(start=ex2_start, asof=ex2_asof, end=ex2_asof), 14)
    Decimal('0.17216108990194')
    >>> round(dcfc_act_act(start=ex3_start, asof=ex3_asof, end=ex3_asof), 14)
    Decimal('1.08243131970956')
    >>> round(dcfc_act_act(start=ex4_start, asof=ex4_asof, end=ex4_asof), 14)
    Decimal('1.32625945055768')
    """
    ## Get all years of interest by checking the leap year:
    years = {year: calendar.isleap(year) for year in range(start.year, asof.year + 1)}

    ## Define the buffer of days for the day count. The former is for non-leap years, the latter for leap years:
    buffer: List[int] = [0, 0]

    ## Iterate over the date range and count:
    for date in _get_date_range(start, asof):
        ## Check the year and modify buffer accordingly:
        if years[date.year]:
            ## Yep, it is a leap year:
            buffer[1] += 1
        else:
            ## Nope, not a leap year:
            buffer[0] += 1

    ## Done, compute and return:
    return Decimal(buffer[0]) / Decimal(365) + Decimal(buffer[1]) / Decimal(366)


@dcc("Act/Act (ICMA)", {"Actual/Actual (ICMA)", "ISMA-99", "Act/Act (ISMA)"})
def dcfc_act_act_icma(start: Date, asof: Date, end: Date, freq: Optional[Decimal] = None) -> Decimal:
    """
    Computes the day count fraction for "Act/Act (ICMA)" convention.

    :param start: The start date of the period.
    :param asof: The date which the day count fraction to be calculated as of.
    :param end: The end date of the period (a.k.a. termination date).
    :return: Day count fraction.

    >>> ex1_start, ex1_asof, ex1_end = datetime.date(2019, 3, 2), datetime.date(2019, 9, 10), datetime.date(2020, 3, 2)
    >>> round(dcfc_act_act_icma(start=ex1_start, asof=ex1_asof, end=ex1_end), 10)
    Decimal('0.5245901639')
    """
    ## Get the number of actual days:
    p1 = Decimal(_get_actual_day_count(start, asof))

    ## Get the number of days in the period:
    p2 = Decimal(_get_actual_day_count(start, end))

    ## Compute the ratio and return:
    return p1 / p2 / Decimal(freq or ONE)


@dcc(
    "Act/360",
    {"Actual/360", "French", "360"},
    _as_ccys({"AUD", "CAD", "CHF", "EUR", "USD", "DKK", "CZK", "HUF", "SEK", "IDR", "NOK", "JPY", "NZD", "THB"}),
)
def dcfc_act_360(start: Date, asof: Date, end: Date, freq: Optional[Decimal] = None) -> Decimal:
    """
    Computes the day count fraction for "Act/360" convention.

    :param start: The start date of the period.
    :param end: The end date of the period.
    :return: Day count fraction.

    >>> ex1_start, ex1_asof = datetime.date(2007, 12, 28), datetime.date(2008, 2, 28)
    >>> ex2_start, ex2_asof = datetime.date(2007, 12, 28), datetime.date(2008, 2, 29)
    >>> ex3_start, ex3_asof = datetime.date(2007, 10, 31), datetime.date(2008, 11, 30)
    >>> ex4_start, ex4_asof = datetime.date(2008, 2, 1), datetime.date(2009, 5, 31)
    >>> round(dcfc_act_360(start=ex1_start, asof=ex1_asof, end=ex1_asof), 14)
    Decimal('0.17222222222222')
    >>> round(dcfc_act_360(start=ex2_start, asof=ex2_asof, end=ex2_asof), 14)
    Decimal('0.17500000000000')
    >>> round(dcfc_act_360(start=ex3_start, asof=ex3_asof, end=ex3_asof), 14)
    Decimal('1.10000000000000')
    >>> round(dcfc_act_360(start=ex4_start, asof=ex4_asof, end=ex4_asof), 14)
    Decimal('1.34722222222222')
    """
    return _get_actual_day_count(start, asof) / Decimal(360)


@dcc("Act/365F", {"Actual/365 Fixed", "English", "365"}, _as_ccys({"GBP", "HKD", "INR", "PLN", "SGD", "ZAR", "MYR"}))
def dcfc_act_365_f(start: Date, asof: Date, end: Date, freq: Optional[Decimal] = None) -> Decimal:
    """
    Computes the day count fraction for the "Act/365F" convention.

    :param start: The start date of the period.
    :param asof: The date which the day count fraction to be calculated as of.
    :param end: The end date of the period (a.k.a. termination date).
    :return: Day count fraction.

    >>> ex1_start, ex1_asof = datetime.date(2007, 12, 28), datetime.date(2008, 2, 28)
    >>> ex2_start, ex2_asof = datetime.date(2007, 12, 28), datetime.date(2008, 2, 29)
    >>> ex3_start, ex3_asof = datetime.date(2007, 10, 31), datetime.date(2008, 11, 30)
    >>> ex4_start, ex4_asof = datetime.date(2008, 2, 1), datetime.date(2009, 5, 31)
    >>> round(dcfc_act_365_f(start=ex1_start, asof=ex1_asof, end=ex1_asof), 14)
    Decimal('0.16986301369863')
    >>> round(dcfc_act_365_f(start=ex2_start, asof=ex2_asof, end=ex2_asof), 14)
    Decimal('0.17260273972603')
    >>> round(dcfc_act_365_f(start=ex3_start, asof=ex3_asof, end=ex3_asof), 14)
    Decimal('1.08493150684932')
    >>> round(dcfc_act_365_f(start=ex4_start, asof=ex4_asof, end=ex4_asof), 14)
    Decimal('1.32876712328767')
    """
    return _get_actual_day_count(start, asof) / Decimal(365)


@dcc("Act/365A", {"Actual/365 Actual"})
def dcfc_act_365_a(start: Date, asof: Date, end: Date, freq: Optional[Decimal] = None) -> Decimal:
    """
    Computes the day count fraction for the "Act/365A" convention.

    :param start: The start date of the period.
    :param asof: The date which the day count fraction to be calculated as of.
    :param end: The end date of the period (a.k.a. termination date).
    :return: Day count fraction.

    >>> ex1_start, ex1_asof = datetime.date(2007, 12, 28), datetime.date(2008, 2, 28)
    >>> ex2_start, ex2_asof = datetime.date(2007, 12, 28), datetime.date(2008, 2, 29)
    >>> ex3_start, ex3_asof = datetime.date(2007, 10, 31), datetime.date(2008, 11, 30)
    >>> ex4_start, ex4_asof = datetime.date(2008, 2, 1), datetime.date(2009, 5, 31)
    >>> round(dcfc_act_365_a(start=ex1_start, asof=ex1_asof, end=ex1_asof), 14)
    Decimal('0.16986301369863')
    >>> round(dcfc_act_365_a(start=ex2_start, asof=ex2_asof, end=ex2_asof), 14)
    Decimal('0.17213114754098')
    >>> round(dcfc_act_365_a(start=ex3_start, asof=ex3_asof, end=ex3_asof), 14)
    Decimal('1.08196721311475')
    >>> round(dcfc_act_365_a(start=ex4_start, asof=ex4_asof, end=ex4_asof), 14)
    Decimal('1.32513661202186')
    """
    return _get_actual_day_count(start, asof) / Decimal(366 if _has_leap_day(start, asof) else 365)


@dcc("Act/365L", {"Actual/365 Leap Year"})
def dcfc_act_365_l(start: Date, asof: Date, end: Date, freq: Optional[Decimal] = None) -> Decimal:
    """
    Computes the day count fraction for the "Act/365L" convention.

    :param start: The start date of the period.
    :param asof: The date which the day count fraction to be calculated as of.
    :param end: The end date of the period (a.k.a. termination date).
    :return: Day count fraction.

    >>> ex1_start, ex1_asof = datetime.date(2007, 12, 28), datetime.date(2008, 2, 28)
    >>> ex2_start, ex2_asof = datetime.date(2007, 12, 28), datetime.date(2008, 2, 29)
    >>> ex3_start, ex3_asof = datetime.date(2007, 10, 31), datetime.date(2008, 11, 30)
    >>> ex4_start, ex4_asof = datetime.date(2008, 2, 1), datetime.date(2009, 5, 31)
    >>> round(dcfc_act_365_l(start=ex1_start, asof=ex1_asof, end=ex1_asof), 14)
    Decimal('0.16939890710383')
    >>> round(dcfc_act_365_l(start=ex2_start, asof=ex2_asof, end=ex2_asof), 14)
    Decimal('0.17213114754098')
    >>> round(dcfc_act_365_l(start=ex3_start, asof=ex3_asof, end=ex3_asof), 14)
    Decimal('1.08196721311475')
    >>> round(dcfc_act_365_l(start=ex4_start, asof=ex4_asof, end=ex4_asof), 14)
    Decimal('1.32876712328767')
    """
    return _get_actual_day_count(start, asof) / Decimal(366 if calendar.isleap(asof.year) else 365)


@dcc("NL/365", {"Actual/365 No Leap Year", "NL365"})
def dcfc_nl_365(start: Date, asof: Date, end: Date, freq: Optional[Decimal] = None) -> Decimal:
    """
    Computes the day count fraction for the "NL/365" convention.

    :param start: The start date of the period.
    :param asof: The date which the day count fraction to be calculated as of.
    :param end: The end date of the period (a.k.a. termination date).
    :return: Day count fraction.

    >>> ex1_start, ex1_asof = datetime.date(2007, 12, 28), datetime.date(2008, 2, 28)
    >>> ex2_start, ex2_asof = datetime.date(2007, 12, 28), datetime.date(2008, 2, 29)
    >>> ex3_start, ex3_asof = datetime.date(2007, 10, 31), datetime.date(2008, 11, 30)
    >>> ex4_start, ex4_asof = datetime.date(2008, 2, 1), datetime.date(2009, 5, 31)
    >>> round(dcfc_nl_365(start=ex1_start, asof=ex1_asof, end=ex1_asof), 14)
    Decimal('0.16986301369863')
    >>> round(dcfc_nl_365(start=ex2_start, asof=ex2_asof, end=ex2_asof), 14)
    Decimal('0.16986301369863')
    >>> round(dcfc_nl_365(start=ex3_start, asof=ex3_asof, end=ex3_asof), 14)
    Decimal('1.08219178082192')
    >>> round(dcfc_nl_365(start=ex4_start, asof=ex4_asof, end=ex4_asof), 14)
    Decimal('1.32602739726027')
    """
    return (_get_actual_day_count(start, asof) - (1 if _has_leap_day(start, asof) else 0)) / Decimal(365)


@dcc("30/360 ISDA", {"30/360 US Municipal", "Bond Basis"})
def dcfc_30_360_isda(start: Date, asof: Date, end: Date, freq: Optional[Decimal] = None) -> Decimal:
    """
    Computes the day count fraction for the "30/360 ISDA" convention.

    :param start: The start date of the period.
    :param asof: The date which the day count fraction to be calculated as of.
    :param end: The end date of the period (a.k.a. termination date).
    :return: Day count fraction.

    >>> ex1_start, ex1_asof = datetime.date(2007, 12, 28), datetime.date(2008, 2, 28)
    >>> ex2_start, ex2_asof = datetime.date(2007, 12, 28), datetime.date(2008, 2, 29)
    >>> ex3_start, ex3_asof = datetime.date(2007, 10, 31), datetime.date(2008, 11, 30)
    >>> ex4_start, ex4_asof = datetime.date(2008, 2, 1), datetime.date(2009, 5, 31)
    >>> round(dcfc_30_360_isda(start=ex1_start, asof=ex1_asof, end=ex1_asof), 14)
    Decimal('0.16666666666667')
    >>> round(dcfc_30_360_isda(start=ex2_start, asof=ex2_asof, end=ex2_asof), 14)
    Decimal('0.16944444444444')
    >>> round(dcfc_30_360_isda(start=ex3_start, asof=ex3_asof, end=ex3_asof), 14)
    Decimal('1.08333333333333')
    >>> round(dcfc_30_360_isda(start=ex4_start, asof=ex4_asof, end=ex4_asof), 14)
    Decimal('1.33333333333333')
    """
    ## Get the new start date, if required:
    if start.day == 31:
        start = datetime.date(start.year, start.month, 30)

    ## Get the new asof date, if required:
    if start.day == 30 and asof.day == 31:
        asof = datetime.date(asof.year, asof.month, 30)

    ## Compute number of days:
    nod = (asof.day - start.day) + 30 * (asof.month - start.month) + 360 * (asof.year - start.year)

    ## Done, compute and return the day count fraction:
    return nod / Decimal(360)


@dcc("30E/360", {"30/360 ISMA", "30/360 European", "30S/360 Special German", "Eurobond Basis"})
def dcfc_30_e_360(start: Date, asof: Date, end: Date, freq: Optional[Decimal] = None) -> Decimal:
    """
    Computes the day count fraction for the "30E/360" convention.

    :param start: The start date of the period.
    :param asof: The date which the day count fraction to be calculated as of.
    :param end: The end date of the period (a.k.a. termination date).
    :return: Day count fraction.

    >>> ex1_start, ex1_asof = datetime.date(2007, 12, 28), datetime.date(2008, 2, 28)
    >>> ex2_start, ex2_asof = datetime.date(2007, 12, 28), datetime.date(2008, 2, 29)
    >>> ex3_start, ex3_asof = datetime.date(2007, 10, 31), datetime.date(2008, 11, 30)
    >>> ex4_start, ex4_asof = datetime.date(2008, 2, 1), datetime.date(2009, 5, 31)
    >>> round(dcfc_30_e_360(start=ex1_start, asof=ex1_asof, end=ex1_asof), 14)
    Decimal('0.16666666666667')
    >>> round(dcfc_30_e_360(start=ex2_start, asof=ex2_asof, end=ex2_asof), 14)
    Decimal('0.16944444444444')
    >>> round(dcfc_30_e_360(start=ex3_start, asof=ex3_asof, end=ex3_asof), 14)
    Decimal('1.08333333333333')
    >>> round(dcfc_30_e_360(start=ex4_start, asof=ex4_asof, end=ex4_asof), 14)
    Decimal('1.33055555555556')
    """
    ## Get the new start date, if required:
    if start.day == 31:
        start = datetime.date(start.year, start.month, 30)

    ## Get the new asof date, if required:
    if asof.day == 31:
        asof = datetime.date(asof.year, asof.month, 30)

    ## Compute number of days:
    nod = (asof.day - start.day) + 30 * (asof.month - start.month) + 360 * (asof.year - start.year)

    ## Done, compute and return the day count fraction:
    return nod / Decimal(360)


@dcc("30E+/360")
def dcfc_30_e_plus_360(start: Date, asof: Date, end: Date, freq: Optional[Decimal] = None) -> Decimal:
    """
    Computes the day count fraction for the "30E+/360" convention.

    :param start: The start date of the period.
    :param asof: The date which the day count fraction to be calculated as of.
    :param end: The end date of the period (a.k.a. termination date).
    :return: Day count fraction.


    >>> ex1_start, ex1_asof = datetime.date(2007, 12, 28), datetime.date(2008, 2, 28)
    >>> ex2_start, ex2_asof = datetime.date(2007, 12, 28), datetime.date(2008, 2, 29)
    >>> ex3_start, ex3_asof = datetime.date(2007, 10, 31), datetime.date(2008, 11, 30)
    >>> ex4_start, ex4_asof = datetime.date(2008, 2, 1), datetime.date(2009, 5, 31)
    >>> round(dcfc_30_e_plus_360(start=ex1_start, asof=ex1_asof, end=ex1_asof), 14)
    Decimal('0.16666666666667')
    >>> round(dcfc_30_e_plus_360(start=ex2_start, asof=ex2_asof, end=ex2_asof), 14)
    Decimal('0.16944444444444')
    >>> round(dcfc_30_e_plus_360(start=ex3_start, asof=ex3_asof, end=ex3_asof), 14)
    Decimal('1.08333333333333')
    >>> round(dcfc_30_e_plus_360(start=ex4_start, asof=ex4_asof, end=ex4_asof), 14)
    Decimal('1.33333333333333')
    """
    ## Get the new start date, if required:
    if start.day == 31:
        start = datetime.date(start.year, start.month, 30)

    ## Get the new asof date, if required:
    if asof.day == 31:
        asof = asof + datetime.timedelta(days=1)

    ## Compute number of days:
    nod = (asof.day - start.day) + 30 * (asof.month - start.month) + 360 * (asof.year - start.year)

    ## Done, compute and return the day count fraction:
    return nod / Decimal(360)


@dcc("30/360 German", {"30E/360 ISDA"})
def dcfc_30_360_german(start: Date, asof: Date, end: Date, freq: Optional[Decimal] = None) -> Decimal:
    """
    Computes the day count fraction.

    :param start: The start date of the period.
    :param asof: The date which the day count fraction to be calculated as of.
    :param end: The end date of the period (a.k.a. termination date).
    :return: Day count fraction.

    >>> ex1_start, ex1_asof = datetime.date(2007, 12, 28), datetime.date(2008, 2, 28)
    >>> ex2_start, ex2_asof = datetime.date(2007, 12, 28), datetime.date(2008, 2, 29)
    >>> ex3_start, ex3_asof = datetime.date(2007, 10, 31), datetime.date(2008, 11, 30)
    >>> ex4_start, ex4_asof = datetime.date(2008, 2, 1), datetime.date(2009, 5, 31)
    >>> round(dcfc_30_360_german(start=ex1_start, asof=ex1_asof, end=ex1_asof), 14)
    Decimal('0.16666666666667')
    >>> round(dcfc_30_360_german(start=ex2_start, asof=ex2_asof, end=ex2_asof), 14)
    Decimal('0.16944444444444')
    >>> round(dcfc_30_360_german(start=ex3_start, asof=ex3_asof, end=ex3_asof), 14)
    Decimal('1.08333333333333')
    >>> round(dcfc_30_360_german(start=ex4_start, asof=ex4_asof, end=ex4_asof), 14)
    Decimal('1.33055555555556')
    """
    ## Get the new start date, if required:
    if start.day == 31 or (start.month == 2 and _is_last_day_of_month(start)):
        d1 = 30
    else:
        d1 = start.day

    ## Get the new asof date, if required:
    if asof.day == 31 or (asof.month == 2 and _is_last_day_of_month(asof) and end != asof):
        d2 = 30
    else:
        d2 = asof.day

    ## Compute number of days:
    nod = (d2 - d1) + 30 * (asof.month - start.month) + 360 * (asof.year - start.year)

    ## Done, compute and return the day count fraction:
    return nod / Decimal(360)


@dcc("30/360 US", {"30U/360", "30US/360"})
def dcfc_30_360_us(start: Date, asof: Date, end: Date, freq: Optional[Decimal] = None) -> Decimal:
    """
    Computes the day count fraction for the "30/360 US" convention.

    :param start: The start date of the period.
    :param asof: The date which the day count fraction to be calculated as of.
    :param end: The end date of the period (a.k.a. termination date).
    :return: Day count fraction.

    >>> ex1_start, ex1_asof = datetime.date(2007, 12, 28), datetime.date(2008, 2, 28)
    >>> ex2_start, ex2_asof = datetime.date(2007, 12, 28), datetime.date(2008, 2, 29)
    >>> ex3_start, ex3_asof = datetime.date(2007, 10, 31), datetime.date(2008, 11, 30)
    >>> ex4_start, ex4_asof = datetime.date(2008, 2, 1), datetime.date(2009, 5, 31)
    >>> round(dcfc_30_360_us(start=ex1_start, asof=ex1_asof, end=ex1_asof), 14)
    Decimal('0.16666666666667')
    >>> round(dcfc_30_360_us(start=ex2_start, asof=ex2_asof, end=ex2_asof), 14)
    Decimal('0.16944444444444')
    >>> round(dcfc_30_360_us(start=ex3_start, asof=ex3_asof, end=ex3_asof), 14)
    Decimal('1.08333333333333')
    >>> round(dcfc_30_360_us(start=ex4_start, asof=ex4_asof, end=ex4_asof), 14)
    Decimal('1.33333333333333')
    """
    ## Get D1 and D2:
    d1 = start.day
    d2 = asof.day

    ## Need to change D1?
    if _is_last_day_of_month(start):
        ## Yep, change it:
        d1 = 30

        ## Shall we change the d2, too?
        if _is_last_day_of_month(asof):
            d2 = 30

    ## Revisit d2:
    if d2 == 31 and (d1 == 30 or d1 == 31):
        d2 = 30

    ## Revisit d1:
    if d1 == 31:
        d1 = 30

    ## Compute number of days:
    nod = (d2 - d1) + 30 * (asof.month - start.month) + 360 * (asof.year - start.year)

    ## Done, return:
    return nod / Decimal(360)



"""
This module provides currency definitions and related functionality.
"""

__all__ = ["Currencies", "Currency", "CurrencyLookupError", "CurrencyRegistry", "CurrencyType"]

from collections import OrderedDict
from dataclasses import dataclass
from decimal import Decimal
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Tuple, Type

from .commons.errors import ProgrammingError
from .commons.numbers import ZERO, MaxPrecisionQuantizer, make_quantizer


class CurrencyLookupError(LookupError):
    """
    Provides a currency lookup error.

    >>> raise CurrencyLookupError("XYZ")
    Traceback (most recent call last):
    ...
    pypara.currencies.CurrencyLookupError: Currency identified by code 'XYZ' does not exist
    """

    def __init__(self, code: str) -> None:
        """
        Initializes a currency lookup error.
        """
        ## Keep the code:
        self.code = code

        ## Set the message:
        super().__init__(f"Currency identified by code '{code}' does not exist")


class CurrencyType(Enum):
    """
    Defines available currency types.
    """

    #: Defines money currency as in common legal tender.
    MONEY = "Money"

    #: Defines precious metals currency type.
    METAL = "Precious Metal"

    #: Defines crypto currency type.
    CRYPTO = "Crypto Currency"

    #: Defines alternative currency type.
    ALTERNATIVE = "Alernative"


@dataclass(frozen=True, order=True)
class Currency:
    """
    Defines currency value object model which is extending ISO 4217 to embrace other currency types.

    Note that you should not call :class:`Currency` constructor directly, but instead use the :method:`Currency.build`.
    :method:`Currency.build` is responsible of performing some checks before creating the currency.

    Try with USD:

    >>> USD = Currency.of("USD", "US Dollars", 2, CurrencyType.MONEY)
    >>> USD.quantize(Decimal("1.005"))
    Decimal('1.00')
    >>> USD.quantize(Decimal("1.015"))
    Decimal('1.02')

    Now, with JPY which has a different precision than USD:

    >>> JPY = Currency.of("JPY", "Japanese Yen", 0, CurrencyType.MONEY)
    >>> JPY.quantize(Decimal("0.5"))
    Decimal('0')
    >>> JPY.quantize(Decimal("1.5"))
    Decimal('2')

    And with a weird currency which has no fixed precision.

    >>> ZZZ = Currency.of("ZZZ", "Some weird currency", -1, CurrencyType.CRYPTO)
    >>> ZZZ.quantize(Decimal("1.0000000000005"))
    Decimal('1.000000000000')
    >>> ZZZ.quantize(Decimal("1.0000000000015"))
    Decimal('1.000000000002')

    Equalities:

    >>> usd1 = Currency.of("USD", "US Dollars", 2, CurrencyType.MONEY)
    >>> usd2 = Currency.of("USD", "US Dollars", 2, CurrencyType.MONEY)
    >>> usdx = Currency.of("USD", "UX Dollars", 2, CurrencyType.MONEY)
    >>> usd1 == usd2
    True
    >>> usd1 == usdx
    False
    >>> hash(usd1) == hash(usd2)
    True
    >>> hash(usd1) == hash(usdx)
    False
    """

    #: Defines the code of the currency.
    code: str

    #: Defines the name of the currency.
    name: str

    #: Defines the number of decimals of the currency.
    decimals: int

    #: Defines the type of the currency.
    type: CurrencyType

    #: Defines the quantiser of the currency.
    quantizer: Decimal

    #: Defines the pre-computed, cached hash.
    hashcache: int

    def __eq__(self, other: Any) -> bool:
        """
        Checks if the `self` and `other` are same currencies.
        """
        return isinstance(other, Currency) and self.hashcache == other.hashcache

    def __hash__(self) -> int:
        """
        Returns the pre-computed and cached hash.
        """
        return self.hashcache

    def quantize(self, qty: Decimal) -> Decimal:
        """
        Quantizes the decimal ``qty`` wrt to ccy's minor units fraction. Note that
        the [ROUND HALF TO EVEN](https://en.wikipedia.org/wiki/Rounding) method
        is used for rounding purposes.

        **Note** that the HALF-TO-EVEN method is inherited from the default decimal context instead of
        explicitly passing it. Therefore, if call-site application is making changes to the default
        context, the rounding method may not be HALF-TO-EVEN anymore.
        """
        return qty.quantize(self.quantizer)

    @classmethod
    def of(cls, code: str, name: str, decimals: int, ctype: CurrencyType) -> "Currency":
        """
        Attempts to create a currency instance and returns it.
        """
        ## Check the code:
        ProgrammingError.passert(isinstance(code, str), "Currency code must be a string")
        ProgrammingError.passert(code.isalpha(), "Currency code must contain only alphabetic characters")
        ProgrammingError.passert(code.isupper(), "Currency code must be all uppercase")

        ## Check the name:
        ProgrammingError.passert(isinstance(name, str), "Currency name must be a string")
        ProgrammingError.passert(name != "", "Currency name can not be empty")
        ProgrammingError.passert(not (name.startswith(" ") or name.endswith(" ")), "Trim the currency name")

        ## Check the decimals:
        ProgrammingError.passert(isinstance(decimals, int), "Number of decimals must be an integer")
        ProgrammingError.passert(decimals >= -1, "Number of decimals can not be less than -1")

        ## Check the type:
        ProgrammingError.passert(isinstance(ctype, CurrencyType), "Currency Type must be of type `CurrencyType`")

        ## Define the quantizer:
        if decimals > 0:
            quantizer = make_quantizer(decimals)
        elif decimals < 0:
            quantizer = MaxPrecisionQuantizer
        else:
            quantizer = ZERO

        ## By now, we should have all required instance attributes. However, we want to compute and cache the hash.
        hashcode = hash((code, name, decimals, ctype, quantizer))

        ## Done, create the currency object and return:
        return Currency(code, name, decimals, ctype, quantizer, hashcode)


class CurrencyRegistry:
    """
    Defines a currency registry model.

    >>> "USD" in Currencies
    True
    >>> Currencies.has("USD")
    True
    >>> "XXX" in Currencies
    False
    >>> Currencies.has("XXX")
    False
    >>> Currencies["USD"].code
    'USD'
    >>> Currencies["USD"].name
    'US Dollar'
    >>> Currencies["USD"].type.name
    'MONEY'
    >>> Currencies.get("USD") == Currencies["USD"]
    True
    >>> Currencies.get("XXX")
    >>> Currencies.get("XXX", default=Currencies["USD"]) == Currencies["USD"]
    True
    >>> assert len(Currencies) == len(Currencies.all)
    >>> assert Currencies.codes == [currency.code for currency in Currencies.all]
    >>> assert Currencies.codenames == [(currency.code, currency.name) for currency in Currencies.all]
    """

    #: Defines the singleton instance.
    __instance = None  # type: CurrencyRegistry

    def __new__(cls) -> "CurrencyRegistry":
        """
        Creates the singleton instance, or returns the existing one.
        """
        ## Do we have the singleton instance?
        if CurrencyRegistry.__instance is None:
            ## Nope, not yet. Creat one:
            CurrencyRegistry.__instance = object.__new__(cls)

        ## Return the singleton instance.
        return CurrencyRegistry.__instance

    def __init__(self) -> None:
        """
        Initializes the currency registry.
        """
        ## Initialize the master registry container.
        self.__registry: Dict[str, Currency] = OrderedDict([])

        ## Initialize the currencies buffer.
        self.__currencies: List[Currency] = []

        ## Initialize the currency codes buffer.
        self.__codes: List[str] = []

        ## Initialize the code/name tuples buffer.
        self.__codenames: List[Tuple[str, str]] = []

        ## Define the registry population context open/close flag.
        self.__ctx_open: bool = False

    def __enter__(self) -> Callable[[Currency], None]:
        """
        Enters the registry population context.
        """
        ## Mark the context as open:
        self.__ctx_open = True

        ## OK, return the add method:
        return self.__register

    def __exit__(self, exc_type: Optional[Type[Exception]], exc_value: Optional[str], tracebackx: Any) -> None:
        """
        Exits the registry population context and performs some finalization tasks.
        """
        ## Re-sort the registry:
        self.__registry = OrderedDict([(c.code, c) for c in sorted(self.__registry.values(), key=lambda x: x.code)])

        ## Re-sort currencies buffer:
        self.__currencies = [c for c in self.__registry.values()]

        ## Re-sort the currency codes buffer:
        self.__codes = [c.code for c in self.__currencies]

        ## Re-sort the choices buffer
        self.__codenames = [(c.code, c.name) for c in self.__currencies]

        ## Close the context:
        self.__ctx_open = False

    def __register(self, currency: Currency) -> None:
        """
        Attempts to add the currency to the registry.

        >>> with Currencies as register:
        ...    register(Currency.of("AED", "UAE Dirham", 2, CurrencyType.MONEY))
        Traceback (most recent call last):
        ...
        ValueError: Currency AED is already registered.
        """
        ## Check of the registry population context is open:
        if not self.__ctx_open:
            ## Nope, raise error:
            raise ProgrammingError("Can not create currencies outside registry context.")

        ## Check if the currency is already added:
        if currency.code in self.__registry:
            raise ValueError(f"Currency {currency.code} is already registered.")

        ## Add to the containers:
        self.__registry[currency.code] = currency

    def __len__(self) -> int:
        """
        Returns the number of registered currencies.
        """
        return len(self.__registry)

    def __contains__(self, code: str) -> bool:
        """
        Checks if a given currency code is available.
        """
        return code in self.__registry

    def __getitem__(self, code: str) -> Currency:
        """
        Returns the currency identified by the code or raises lookup error.

        >>> Currencies["USD"].code
        'USD'
        >>> Currencies["NON-EXISTING"].code
        Traceback (most recent call last):
        ...
        pypara.currencies.CurrencyLookupError: Currency identified by code 'NON-EXISTING' does not exist
        """
        try:
            return self.__registry[code]
        except KeyError:
            raise CurrencyLookupError(code)

    def has(self, code: str) -> bool:
        """
        Indicates if the code is a valid currency code.
        """
        return code in self.__registry

    def get(self, code: str, default: Optional[Currency] = None) -> Optional[Currency]:
        """
        Returns the currency for the given code.

        Note that if the code is not a valid currency code, a currency lookup error is raised.
        """
        return self.__registry.get(code, default)

    @property
    def all(self) -> List["Currency"]:
        """
        Returns the list of currencies.
        """
        return self.__currencies

    @property
    def codes(self) -> List[str]:
        """
        Returns a list of codes.
        """
        return self.__codes

    @property
    def codenames(self) -> List[Tuple[str, str]]:
        """
        Returns a list of code/name tuples.
        """
        return self.__codenames


#: Defines the global currencies registry.
Currencies = CurrencyRegistry()


## Create and register currencies in one go:
with Currencies as register:
    register(Currency.of("AED", "UAE Dirham", 2, CurrencyType.MONEY))
    register(Currency.of("AFN", "Afghani", 2, CurrencyType.MONEY))
    register(Currency.of("ALL", "Lek", 2, CurrencyType.MONEY))
    register(Currency.of("AMD", "Armenian Dram", 2, CurrencyType.MONEY))
    register(Currency.of("ANG", "Netherlands Antillean Guilder", 2, CurrencyType.MONEY))
    register(Currency.of("AOA", "Kwanza", 2, CurrencyType.MONEY))
    register(Currency.of("ARS", "Argentine Peso", 2, CurrencyType.MONEY))
    register(Currency.of("AUD", "Australian Dollar", 2, CurrencyType.MONEY))
    register(Currency.of("AWG", "Aruban Florin", 2, CurrencyType.MONEY))
    register(Currency.of("AZN", "Azerbaijanian Manat", 2, CurrencyType.MONEY))
    register(Currency.of("BAM", "Convertible Mark", 2, CurrencyType.MONEY))
    register(Currency.of("BBD", "Barbados Dollar", 2, CurrencyType.MONEY))
    register(Currency.of("BCH", "Bitcoin Cash", -1, CurrencyType.CRYPTO))
    register(Currency.of("BDT", "Taka", 2, CurrencyType.MONEY))
    register(Currency.of("BGN", "Bulgarian Lev", 2, CurrencyType.MONEY))
    register(Currency.of("BHD", "Bahraini Dinar", 3, CurrencyType.MONEY))
    register(Currency.of("BIF", "Burundi Franc", 0, CurrencyType.MONEY))
    register(Currency.of("BMD", "Bermudian Dollar", 2, CurrencyType.MONEY))
    register(Currency.of("BND", "Brunei Dollar", 2, CurrencyType.MONEY))
    register(Currency.of("BOB", "Boliviano", 2, CurrencyType.MONEY))
    register(Currency.of("BOV", "Mvdol", 2, CurrencyType.MONEY))
    register(Currency.of("BRL", "Brazilian Real", 2, CurrencyType.MONEY))
    register(Currency.of("BSD", "Bahamian Dollar", 2, CurrencyType.MONEY))
    register(Currency.of("BTC", "Bitcoin", -1, CurrencyType.CRYPTO))
    register(Currency.of("BTN", "Ngultrum", 2, CurrencyType.MONEY))
    register(Currency.of("BWP", "Pula", 2, CurrencyType.MONEY))
    register(Currency.of("BYR", "Belarussian Ruble", 0, CurrencyType.MONEY))
    register(Currency.of("BZD", "Belize Dollar", 2, CurrencyType.MONEY))
    register(Currency.of("CAD", "Canadian Dollar", 2, CurrencyType.MONEY))
    register(Currency.of("CDF", "Congolese Franc", 2, CurrencyType.MONEY))
    register(Currency.of("CHE", "WIR Euro", 2, CurrencyType.MONEY))
    register(Currency.of("CHF", "Swiss Franc", 2, CurrencyType.MONEY))
    register(Currency.of("CHW", "WIR Franc", 2, CurrencyType.MONEY))
    register(Currency.of("CLF", "Unidad de Fomento", 4, CurrencyType.MONEY))
    register(Currency.of("CLP", "Chilean Peso", 0, CurrencyType.MONEY))
    register(Currency.of("CNH", "Yuan Renminbi (Off-shore)", 2, CurrencyType.MONEY))
    register(Currency.of("CNY", "Yuan Renminbi", 2, CurrencyType.MONEY))
    register(Currency.of("COP", "Colombian Peso", 2, CurrencyType.MONEY))
    register(Currency.of("COU", "Unidad de Valor Real", 2, CurrencyType.MONEY))
    register(Currency.of("CRC", "Costa Rican Colon", 2, CurrencyType.MONEY))
    register(Currency.of("CUC", "Peso Convertible", 2, CurrencyType.MONEY))
    register(Currency.of("CUP", "Cuban Peso", 2, CurrencyType.MONEY))
    register(Currency.of("CVE", "Cabo Verde Escudo", 2, CurrencyType.MONEY))
    register(Currency.of("CZK", "Czech Koruna", 2, CurrencyType.MONEY))
    register(Currency.of("DASH", "Dash", -1, CurrencyType.CRYPTO))
    register(Currency.of("DJF", "Djibouti Franc", 0, CurrencyType.MONEY))
    register(Currency.of("DKK", "Danish Krone", 2, CurrencyType.MONEY))
    register(Currency.of("DOP", "Dominican Peso", 2, CurrencyType.MONEY))
    register(Currency.of("DZD", "Algerian Dinar", 2, CurrencyType.MONEY))
    register(Currency.of("EGP", "Egyptian Pound", 2, CurrencyType.MONEY))
    register(Currency.of("EOS", "EOSIO", -1, CurrencyType.CRYPTO))
    register(Currency.of("ERN", "Nakfa", 2, CurrencyType.MONEY))
    register(Currency.of("ETB", "Ethiopian Birr", 2, CurrencyType.MONEY))
    register(Currency.of("ETC", "Ethereum Classic", -1, CurrencyType.CRYPTO))
    register(Currency.of("ETH", "Ethereum", -1, CurrencyType.CRYPTO))
    register(Currency.of("EUR", "Euro", 2, CurrencyType.MONEY))
    register(Currency.of("FJD", "Fiji Dollar", 2, CurrencyType.MONEY))
    register(Currency.of("FKP", "Falkland Islands Pound", 2, CurrencyType.MONEY))
    register(Currency.of("GBP", "Pound Sterling", 2, CurrencyType.MONEY))
    register(Currency.of("GEL", "Lari", 2, CurrencyType.MONEY))
    register(Currency.of("GHS", "Ghana Cedi", 2, CurrencyType.MONEY))
    register(Currency.of("GIP", "Gibraltar Pound", 2, CurrencyType.MONEY))
    register(Currency.of("GMD", "Dalasi", 2, CurrencyType.MONEY))
    register(Currency.of("GNF", "Guinea Franc", 0, CurrencyType.MONEY))
    register(Currency.of("GTQ", "Quetzal", 2, CurrencyType.MONEY))
    register(Currency.of("GYD", "Guyana Dollar", 2, CurrencyType.MONEY))
    register(Currency.of("HKD", "Hong Kong Dollar", 2, CurrencyType.MONEY))
    register(Currency.of("HNL", "Lempira", 2, CurrencyType.MONEY))
    register(Currency.of("HRK", "Kuna", 2, CurrencyType.MONEY))
    register(Currency.of("HTG", "Gourde", 2, CurrencyType.MONEY))
    register(Currency.of("HUF", "Forint", 2, CurrencyType.MONEY))
    register(Currency.of("IDR", "Rupiah", 2, CurrencyType.MONEY))
    register(Currency.of("ILS", "New Israeli Sheqel", 2, CurrencyType.MONEY))
    register(Currency.of("INR", "Indian Rupee", 2, CurrencyType.MONEY))
    register(Currency.of("IOT", "IOTA", -1, CurrencyType.CRYPTO))
    register(Currency.of("IQD", "Iraqi Dinar", 3, CurrencyType.MONEY))
    register(Currency.of("IRR", "Iranian Rial", 2, CurrencyType.MONEY))
    register(Currency.of("ISK", "Iceland Krona", 0, CurrencyType.MONEY))
    register(Currency.of("JMD", "Jamaican Dollar", 2, CurrencyType.MONEY))
    register(Currency.of("JOD", "Jordanian Dinar", 3, CurrencyType.MONEY))
    register(Currency.of("JPY", "Yen", 0, CurrencyType.MONEY))
    register(Currency.of("KES", "Kenyan Shilling", 2, CurrencyType.MONEY))
    register(Currency.of("KGS", "Som", 2, CurrencyType.MONEY))
    register(Currency.of("KHR", "Riel", 2, CurrencyType.MONEY))
    register(Currency.of("KMF", "Comoro Franc", 0, CurrencyType.MONEY))
    register(Currency.of("KPW", "North Korean Won", 2, CurrencyType.MONEY))
    register(Currency.of("KRW", "Won", 0, CurrencyType.MONEY))
    register(Currency.of("KWD", "Kuwaiti Dinar", 3, CurrencyType.MONEY))
    register(Currency.of("KYD", "Cayman Islands Dollar", 2, CurrencyType.MONEY))
    register(Currency.of("KZT", "Tenge", 2, CurrencyType.MONEY))
    register(Currency.of("LAK", "Kip", 2, CurrencyType.MONEY))
    register(Currency.of("LBP", "Lebanese Pound", 2, CurrencyType.MONEY))
    register(Currency.of("LINK", "Chainlink", -1, CurrencyType.CRYPTO))
    register(Currency.of("LKR", "Sri Lanka Rupee", 2, CurrencyType.MONEY))
    register(Currency.of("LRD", "Liberian Dollar", 2, CurrencyType.MONEY))
    register(Currency.of("LSL", "Loti", 2, CurrencyType.MONEY))
    register(Currency.of("LTC", "Litecoin", -1, CurrencyType.CRYPTO))
    register(Currency.of("LYD", "Libyan Dinar", 3, CurrencyType.MONEY))
    register(Currency.of("MAD", "Moroccan Dirham", 2, CurrencyType.MONEY))
    register(Currency.of("MDL", "Moldovan Leu", 2, CurrencyType.MONEY))
    register(Currency.of("MGA", "Malagasy Ariary", 2, CurrencyType.MONEY))
    register(Currency.of("MKD", "Denar", 2, CurrencyType.MONEY))
    register(Currency.of("MMK", "Kyat", 2, CurrencyType.MONEY))
    register(Currency.of("MNT", "Tugrik", 2, CurrencyType.MONEY))
    register(Currency.of("MOP", "Pataca", 2, CurrencyType.MONEY))
    register(Currency.of("MRO", "Ouguiya", 2, CurrencyType.MONEY))
    register(Currency.of("MUR", "Mauritius Rupee", 2, CurrencyType.MONEY))
    register(Currency.of("MVR", "Rufiyaa", 2, CurrencyType.MONEY))
    register(Currency.of("MWK", "Kwacha", 2, CurrencyType.MONEY))
    register(Currency.of("MXN", "Mexican Peso", 2, CurrencyType.MONEY))
    register(Currency.of("MXV", "Mexican Unidad de Inversion (UDI)", 2, CurrencyType.MONEY))
    register(Currency.of("MYR", "Malaysian Ringgit", 2, CurrencyType.MONEY))
    register(Currency.of("MZN", "Mozambique Metical", 2, CurrencyType.MONEY))
    register(Currency.of("NAD", "Namibia Dollar", 2, CurrencyType.MONEY))
    register(Currency.of("NEO", "NEO", -1, CurrencyType.CRYPTO))
    register(Currency.of("NGN", "Naira", 2, CurrencyType.MONEY))
    register(Currency.of("NIO", "Cordoba Oro", 2, CurrencyType.MONEY))
    register(Currency.of("NOK", "Norwegian Krone", 2, CurrencyType.MONEY))
    register(Currency.of("NPR", "Nepalese Rupee", 2, CurrencyType.MONEY))
    register(Currency.of("NZD", "New Zealand Dollar", 2, CurrencyType.MONEY))
    register(Currency.of("OMG", "OmiseGO", -1, CurrencyType.CRYPTO))
    register(Currency.of("OMR", "Rial Omani", 3, CurrencyType.MONEY))
    register(Currency.of("PAB", "Balboa", 2, CurrencyType.MONEY))
    register(Currency.of("PEN", "Nuevo Sol", 2, CurrencyType.MONEY))
    register(Currency.of("PGK", "Kina", 2, CurrencyType.MONEY))
    register(Currency.of("PHP", "Philippine Peso", 2, CurrencyType.MONEY))
    register(Currency.of("PKR", "Pakistan Rupee", 2, CurrencyType.MONEY))
    register(Currency.of("PLN", "Zloty", 2, CurrencyType.MONEY))
    register(Currency.of("PYG", "Guarani", 0, CurrencyType.MONEY))
    register(Currency.of("QAR", "Qatari Rial", 2, CurrencyType.MONEY))
    register(Currency.of("RON", "Romanian Leu", 2, CurrencyType.MONEY))
    register(Currency.of("RSD", "Serbian Dinar", 2, CurrencyType.MONEY))
    register(Currency.of("RUB", "Russian Ruble", 2, CurrencyType.MONEY))
    register(Currency.of("RWF", "Rwanda Franc", 0, CurrencyType.MONEY))
    register(Currency.of("SAR", "Saudi Riyal", 2, CurrencyType.MONEY))
    register(Currency.of("SBD", "Solomon Islands Dollar", 2, CurrencyType.MONEY))
    register(Currency.of("SCR", "Seychelles Rupee", 2, CurrencyType.MONEY))
    register(Currency.of("SDG", "Sudanese Pound", 2, CurrencyType.MONEY))
    register(Currency.of("SEK", "Swedish Krona", 2, CurrencyType.MONEY))
    register(Currency.of("SGD", "Singapore Dollar", 2, CurrencyType.MONEY))
    register(Currency.of("SHP", "Saint Helena Pound", 2, CurrencyType.MONEY))
    register(Currency.of("SLL", "Leone", 2, CurrencyType.MONEY))
    register(Currency.of("SOS", "Somali Shilling", 2, CurrencyType.MONEY))
    register(Currency.of("SRD", "Surinam Dollar", 2, CurrencyType.MONEY))
    register(Currency.of("SSP", "South Sudanese Pound", 2, CurrencyType.MONEY))
    register(Currency.of("STD", "Dobra", 2, CurrencyType.MONEY))
    register(Currency.of("SVC", "El Salvador Colon", 2, CurrencyType.MONEY))
    register(Currency.of("SYP", "Syrian Pound", 2, CurrencyType.MONEY))
    register(Currency.of("SZL", "Lilangeni", 2, CurrencyType.MONEY))
    register(Currency.of("THB", "Baht", 2, CurrencyType.MONEY))
    register(Currency.of("TJS", "Somoni", 2, CurrencyType.MONEY))
    register(Currency.of("TMT", "Turkmenistan New Manat", 2, CurrencyType.MONEY))
    register(Currency.of("TND", "Tunisian Dinar", 3, CurrencyType.MONEY))
    register(Currency.of("TOP", "Pa'anga", 2, CurrencyType.MONEY))
    register(Currency.of("TRY", "Turkish Lira", 2, CurrencyType.MONEY))
    register(Currency.of("TTD", "Trinidad and Tobago Dollar", 2, CurrencyType.MONEY))
    register(Currency.of("TWD", "New Taiwan Dollar", 2, CurrencyType.MONEY))
    register(Currency.of("TZS", "Tanzanian Shilling", 2, CurrencyType.MONEY))
    register(Currency.of("UAH", "Hryvnia", 2, CurrencyType.MONEY))
    register(Currency.of("UGX", "Uganda Shilling", 0, CurrencyType.MONEY))
    register(Currency.of("USD", "US Dollar", 2, CurrencyType.MONEY))
    register(Currency.of("USN", "US Dollar (Next day)", 2, CurrencyType.MONEY))
    register(Currency.of("UYI", "Uruguay Peso en Unidades Indexadas", 0, CurrencyType.MONEY))
    register(Currency.of("UYU", "Peso Uruguayo", 2, CurrencyType.MONEY))
    register(Currency.of("UZS", "Uzbekistan Sum", 2, CurrencyType.MONEY))
    register(Currency.of("VEF", "Bolivar", 2, CurrencyType.MONEY))
    register(Currency.of("VND", "Dong", 0, CurrencyType.MONEY))
    register(Currency.of("VUV", "Vatu", 0, CurrencyType.MONEY))
    register(Currency.of("WST", "Tala", 2, CurrencyType.MONEY))
    register(Currency.of("XAF", "Central African CFA Franc BCEAO", 2, CurrencyType.MONEY))
    register(Currency.of("XAG", "Silver", -1, CurrencyType.METAL))
    register(Currency.of("XAU", "Gold", -1, CurrencyType.METAL))
    register(Currency.of("XCD", "East Caribbean Dollar", 2, CurrencyType.MONEY))
    register(Currency.of("XLM", "Stellar", -1, CurrencyType.CRYPTO))
    register(Currency.of("XMR", "Monero", -1, CurrencyType.CRYPTO))
    register(Currency.of("XOF", "West African CFA Franc BCEAO", 2, CurrencyType.MONEY))
    register(Currency.of("XPD", "Palladium", -1, CurrencyType.METAL))
    register(Currency.of("XPT", "Platinum", -1, CurrencyType.METAL))
    register(Currency.of("XRP", "Ripple", -1, CurrencyType.CRYPTO))
    register(Currency.of("XSU", "Sucre", -1, CurrencyType.MONEY))
    register(Currency.of("XUA", "ADB Unit of Account", -1, CurrencyType.MONEY))
    register(Currency.of("YER", "Yemeni Rial", 2, CurrencyType.MONEY))
    register(Currency.of("ZAR", "Rand", 2, CurrencyType.MONEY))
    register(Currency.of("ZEC", "Zcash", -1, CurrencyType.CRYPTO))
    register(Currency.of("ZMW", "Zambian Kwacha", 2, CurrencyType.MONEY))
    register(Currency.of("ZWL", "Zimbabwe Dollar", 2, CurrencyType.MONEY))



"""
This module provides account types, accounts and a chart of accounts model definitions.

Account's are identified by unique :py:class:`Code` instances which is a :py:class:`typing.NewType` of :py:class:`str`.

A chart of accounts (:py:class:`COA`) is initialized with 5 core accounts of types:

#. :py:attr:`AccountType.ASSETS`
#. :py:attr:`AccountType.LIABILITIES`
#. :py:attr:`AccountType.EQUITIES`
#. :py:attr:`AccountType.REVENUES`
#. :py:attr:`AccountType.EXPENSES`

A simple chart of accounts can be initialized as follows:

>>> coa = COA()
>>> for code, acct in coa:
...     print((code, acct.name))
('1', 'Assets')
('2', 'Liabilities')
('3', 'Equities')
('4', 'Revenues')
('5', 'Expenses')

We can retrieve accounts by account code:

>>> coa.find(Code("1")).name
'Assets'
>>> coa.find(Code("2")).name
'Liabilities'
>>> coa.find(Code("3")).name
'Equities'
>>> coa.find(Code("4")).name
'Revenues'
>>> coa.find(Code("5")).name
'Expenses'
>>> coa.find(Code("boguscode")) is None
True

We can add further accounts:

>>> liquidity = coa.add(Code("1"), Code("1000"), "Liquidity")
>>> bankaccnt = coa.add(liquidity.code, Code("1001"), "Bank Account")
>>> for code, acct in coa:
...     print((code, acct.name))
('1', 'Assets')
('2', 'Liabilities')
('3', 'Equities')
('4', 'Revenues')
('5', 'Expenses')
('1000', 'Liquidity')
('1001', 'Bank Account')
>>> coa.find(Code("1001")).parent.name
'Liquidity'
>>> coa.find(Code("1001")).name
'Bank Account'

We can also print the chart of accounts in a tree-like structure.

>>> coa.print()
[1] Assets
    [1000] Liquidity
        [1001] Bank Account
[2] Liabilities
[3] Equities
[4] Revenues
[5] Expenses
"""

__all__ = [
    "Account",
    "AccountType",
    "COA",
    "Code",
    "ReadChartOfAccounts",
]

from abc import abstractmethod
from collections import OrderedDict
from dataclasses import InitVar, dataclass, field
from enum import Enum
from typing import Dict, Iterable, Iterator, List, NewType, Optional, Protocol, Tuple, runtime_checkable

#: Defines a new-type for account codes.
Code = NewType("Code", str)


class AccountType(Enum):
    """
    Provides an enumeration of types of accounts in balance sheet and income statement.
    """

    #: Indicates asset accounts related to economic resources which are beneficial to the entity.
    ASSETS = "ASSETS"

    #: Indicates liability accounts related to debts and future obligations to the entity.
    LIABILITIES = "LIABILITIES"

    #: Indicates equity accounts related to claims of owners and/or shareholders of the entity.
    EQUITIES = "EQUITIES"

    #: Indicates accounts for revenues resulting in increase in :py:attr:`AccountType.EQUITIES`.
    REVENUES = "REVENUES"

    #: Indicates accounts for expenses resulting in decrease in :py:attr:`AccountType.EQUITIES`.
    EXPENSES = "EXPENSES"

    def __lt__(self, other: "AccountType") -> bool:
        """
        Provides a comparison for account type enums.
        """
        ## Provides the order of enums:
        order: Dict["AccountType", int] = {
            AccountType.ASSETS: 0,
            AccountType.LIABILITIES: 1,
            AccountType.EQUITIES: 2,
            AccountType.REVENUES: 3,
            AccountType.EXPENSES: 4,
        }

        ## Compare and return the order of enums:
        return order[self] < order[other]


@runtime_checkable
class Account(Protocol):
    """
    Provides a base account model.
    """

    @property
    @abstractmethod
    def code(self) -> Code:
        """
        Code of the account.
        """
        ...

    @property
    @abstractmethod
    def name(self) -> str:
        """
        Name of the account.
        """
        ...

    @property
    @abstractmethod
    def type(self) -> AccountType:
        """
        Type of the account.
        """
        ...

    @property
    @abstractmethod
    def coa(self) -> "COA":
        """
        Chart of accounts of the account.
        """
        ...

    @property
    @abstractmethod
    def parent(self) -> Optional["Account"]:
        """
        Parent of the account, if any.
        """
        ...


@dataclass(frozen=True)
class RootAccount:
    """
    Provides root account model.
    """

    #: Code of the account.
    code: Code

    #: Name of the account.
    name: str

    #: Type of the account.
    type: AccountType

    #: Chart of accounts the account belongs to.
    coa: "COA"

    @property
    def parent(self) -> Optional["Account"]:
        """
        Parent account.
        """
        ## I wish I could do `parent: ClassVar[Optional["Account"]] = None` instead!
        return None


@dataclass(frozen=True)
class SubAccount:
    """
    Provides sub-account model.
    """

    #: Code of the account.
    code: Code

    #: Name of the account.
    name: str

    #: Parent account.
    parent: "Account"

    @property
    def type(self) -> AccountType:
        """
        Type of the account.
        """
        return self.parent.type

    @property
    def coa(self) -> "COA":
        """
        Chart of accounts the account belongs to.
        """
        return self.parent.coa


@dataclass(frozen=True)
class COA:
    """
    Provides a model of chart of accounts.
    """

    @dataclass
    class Node:
        """
        Provides a node definition for the chart of accounts tree.
        """

        #: Account of the node.
        account: Account

        #: Sub-accounts of the account of the node.
        children: List["COA.Node"]

    #: Internal buffer for accounts.
    _accounts: Dict[Code, Account] = field(default_factory=OrderedDict, hash=False)

    #: Internal buffer for sub-accounts.
    _subaccounts: Dict[Account, List[Account]] = field(default_factory=OrderedDict, hash=False)

    #: Specification to initialize root accounts.
    rootspec: InitVar[Optional[Dict[AccountType, Tuple[Code, str]]]] = None

    def __post_init__(self, rootspec: Optional[Dict[AccountType, Tuple[Code, str]]]) -> None:
        """
        Initializes the root accounts buffer.
        """
        ## Do we have a rootspec?
        rootspec = rootspec or {}

        ## Iterate over account types and initialize root accounts:
        for c, t in enumerate(AccountType, start=1):
            ## Attempt to get or initialize code/name tuple for the type:
            code, name = rootspec.get(t, (Code(str(c)), t.name.capitalize()))

            ## Create the account and add to the buffer:
            self._accounts[code] = RootAccount(code, name, t, self)

    def __iter__(self) -> Iterator[Tuple[Code, Account]]:
        """
        Returns an iterable of account code and account tuples.
        """
        return ((c, a) for c, a in self._accounts.items())

    @property
    def accounts(self) -> Iterable[Account]:
        """
        All accounts of the chart of accounts.
        """
        return iter(self._accounts.values())

    @property
    def toplevel(self) -> Iterable[Account]:
        """
        Top-level accounts (balance sheet and income statement accounts) of the chart of accounts.
        """
        return (a for a in self.accounts if a.parent is None)

    @property
    def structure(self) -> Iterable["COA.Node"]:
        """
        Tree-like structure of the chart of accounts.
        """
        return map(self.nodify, self.toplevel)

    def find(self, code: Code) -> Optional[Account]:
        """
        Attempts to find and return the account by the given code.

        :param code: Code of the account we want to retrieve.
        :return: :py:class:`Account` identified by the given code if found, ``None`` otherwise.
        """
        return self._accounts.get(code, None)

    def subaccounts(self, account: Account) -> List[Account]:
        """
        Attempts to find and return sub-accounts of the given account.

        :param account: Account we want to retrieve sub-accounts of.
        :return: List of sub-accounts.
        """
        return self._subaccounts.get(account, [])

    def nodify(self, account: Account) -> "COA.Node":
        """
        Compiles a :py:class:`Node` instance for the given account.

        :param account: Account we want to compile the :py:class:`Node` for.
        :return: Account :py:class:`Node` within the chart of accounts.
        """
        return self.Node(account, [self.nodify(a) for a in self.subaccounts(account)])

    def add(self, parent: Code, code: Code, name: str) -> Account:
        """
        Attempts to add a new account to the chart of accounts.

        :param parent: Code of the parent account.
        :param code: Account code.
        :param name: Account name.
        :return: Newly created account (or existing one).
        """
        ## Check if parent and code are same:
        if parent == code:
            raise ValueError("An account can not be the parent of itself.")

        ## Attempt to get the parent instance:
        parentinstance = self._accounts.get(parent)

        ## Check if we have a parent instance:
        if parentinstance is None:
            raise ValueError("Parent account is not (yet) defined.")

        ## Check if we already have an account:
        if code in self._accounts:
            ## Get the account:
            account = self._accounts[code]

            ## Check account information is consistent:
            if account.parent == parentinstance and account.name == name and account.code == code:
                return account
            else:
                raise ValueError("Account name, code and parent do not match existing chart of accounts member.")

        ## Create the account:
        account = SubAccount(code, name, self._accounts[parent])

        ## Add to the COA:
        self._accounts[code] = account

        ## Add the account to children buffer:
        if account.parent not in self._subaccounts:
            self._subaccounts[account.parent] = []
        self._subaccounts[account.parent].append(account)

        ## Done, return the new account:
        return account

    def print(self) -> None:
        """
        Prints the chart of accounts in a tree-like form.

        :param coa: Chart of accounts to print.
        """
        for tree in self.structure:
            self._print_node(tree, 0)

    @classmethod
    def _print_node(cls, node: "COA.Node", level: int = 0) -> None:
        """
        Prints a node.

        This is an auxiliary function to :py:func:`print_coa`.

        :param node: Node to print.
        :param level: Level to print node at.
        """
        print(f"{''.join(['    '] * level)}[{node.account.code}] {node.account.name}")
        for c in node.children:
            cls._print_node(c, level + 1)


class ReadChartOfAccounts(Protocol):
    """
    Type of functions which read chart-of-accounts from a source.
    """

    def __call__(self) -> COA:
        pass



"""
This module provides the setup procedure.
"""

import os
import re

from setuptools import find_packages, setup

setup(
    name="pypara",
    version=re.search(r"__version__\s*=\s*['\"]([^'\"]*)['\"]", open("pypara/__init__.py").read()).group(1),
    description="Currencies, Monetary Value Arithmetic/Conversion and Some Type Convenience",
    long_description=open(os.path.join(os.path.abspath(os.path.dirname(__file__)), "README.md")).read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    author="Vehbi Sinan Tunalioglu",
    author_email="vst@vsthost.com",
    url="https://github.com/vst/pypara",
    license="BSD",
    package_data={"pypara": ["py.typed"]},
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=["python-dateutil<3.0"],
    dependency_links=[],
    scripts=[],
)



"""
This module provides common numeric type definitions, constants and functions.
"""

__all__ = [
    "Amount",
    "CENT",
    "DecimalLike",
    "HUNDRED",
    "isum",
    "make_quantizer",
    "MaxPrecision",
    "MaxPrecisionQuantizer",
    "NaturalNumber",
    "Numeric",
    "ONE",
    "PositiveInteger",
    "Quantity",
    "quantize12",
    "quantize2",
    "quantize4",
    "quantize8",
    "Quantizer12",
    "Quantizer2",
    "Quantizer4",
    "Quantizer8",
    "sign",
    "ZERO",
]

import sys
from decimal import Decimal
from typing import Callable, Iterable, NewType, Optional, TypeVar, cast


class NaturalNumber(int):
    """
    Provides a *newtype* for natural numbers ``[0, 1, ...)``.

    Call-site should ensure that the constructor does not throw assertion errors.
    """

    def __new__(cls, value: int) -> "NaturalNumber":
        """
        Attempts to create a new instance.

        :param value: A non-negative integer.
        :raises AssertionError: If ``value`` is not a non-negative integer.
        """
        assert value >= 0
        return int(value)  # type: ignore


class PositiveInteger(int):
    """
    Provides a *newtype* for positive integers ``[1, ...)``.

    Call-site should ensure that the constructor does not throw assertion errors.
    """

    def __new__(cls, value: int) -> "PositiveInteger":
        """
        Attempts to create a new instance.

        :param value: A positive integer.
        :raises AssertionError: If ``value`` is not a positive integer.
        """
        assert value > 0
        return int(value)  # type: ignore


#: Defines a new-type for non-negative (absolute, unsigned)
#: :py:class:`decimal.Decimal` values as *amount*.
#:
#: This is just a convenience for adding further semantics to type signatures.
Amount = NewType("Amount", Decimal)

#: Defines a new-type for :py:class:`decimal.Decimal` values as *quantities*.
#:
#: This is just a convenience for adding further semantics to type signatures.
Quantity = NewType("Quantity", Decimal)

#: Defines a type alias for numeric values.
Numeric = TypeVar("Numeric", int, float, Decimal, Amount, Quantity)

#: Defines a type variable for values of type either :py:class:`decimal.Decimal`
#: or a sub-class or new-type based on : py:class:`Decimal` such as
#: :py:class:`Amount` and a :py:class:`Quantity`.
DecimalLike = TypeVar("DecimalLike", bound=Decimal)

#: Defines the constant for :py:class:`decimal.Decimal` value ``0``.
ZERO = Decimal("0")

#: Defines the constant for :py:class:`decimal.Decimal` value ``0.01``.
CENT = Decimal("0.01")

#: Defines the constant for :py:class:`decimal.Decimal` value ``1``.
ONE = Decimal("1")

#: Defines the constant for :py:class:`decimal.Decimal` value ``100``.
HUNDRED = Decimal("100")


def make_quantizer(precision: int) -> Decimal:
    """
    Creates a quantifier as per the given precision.
    """
    return Decimal(f"0.{''.join(['0' * precision])}")


def make_quantize_func(quantizer: Decimal) -> Callable[[Decimal], Decimal]:
    """
    Creates a quantize function with the given quantizer.
    """
    return lambda x: x.quantize(quantizer)


#: Defines the maximum precision of the monetary values and operations in this library.
MaxPrecision: int = 12

#: Defines the maximum precision quantifier.
MaxPrecisionQuantizer: Decimal = make_quantizer(MaxPrecision)

#: Defines a quantizer with 2 decimals.
Quantizer2 = make_quantizer(2)

#: Defines a quantizer with 4 decimals.
Quantizer4 = make_quantizer(4)

#: Defines a quantizer with 8 decimals.
Quantizer8 = make_quantizer(8)

#: Defines a quantizer with 12 decimals.
Quantizer12 = make_quantizer(12)


quantize2 = make_quantize_func(Quantizer2)
quantize2.__doc__ = """
Quantizes the given decimal (by 2 decimals).

>>> quantize2(Decimal("0.005"))
Decimal('0.00')
>>> quantize2(Decimal("0.015"))
Decimal('0.02')
"""


quantize4 = make_quantize_func(Quantizer4)
quantize4.__doc__ = """
Quantizes the given decimal (by 4 decimals).

>>> quantize4(Decimal("0.00005"))
Decimal('0.0000')
>>> quantize4(Decimal("0.00015"))
Decimal('0.0002')
"""

quantize8 = make_quantize_func(Quantizer8)
quantize8.__doc__ = """
Quantizes the given decimal (by 8 decimals).

>>> quantize8(Decimal("0.000000005"))
Decimal('0E-8')
>>> quantize8(Decimal("0.000000015"))
Decimal('2E-8')
"""


quantize12 = make_quantize_func(Quantizer12)
quantize12.__doc__ = """
Quantizes the given decimal (by 12 decimals).

>>> quantize12(Decimal("0.0000000000005"))
Decimal('0E-12')
>>> quantize12(Decimal("0.0000000000015"))
Decimal('2E-12')
"""


def isum(xs: Iterable[DecimalLike], start: Optional[DecimalLike] = None) -> DecimalLike:
    """
    Computes the sum of an iterable of :py:class:`DecimalLike` values such as :py:class:`Amount` or
    :py:class:`Quantity` including :py:class:`Decimal` itself.

    The return type is the same as the input element type. The base condition is :py:const:`ZERO` of
    :py:class:`decimal.Decimal` type but cast to the type variable if required.

    :param xs: An iterable of :py:class:`Decimal`-like values.
    :param start: Optional initial value. This defaults to :py:const:`ZERO` in the implementation.
    :return: Sum of the elements in the same type as the elements in the argument.

    >>> isum([Amount(ONE), Amount(ONE)])  # Return value is of type `Amount` during type-checking.
    Decimal('2')
    >>> isum([Quantity(ONE), Quantity(ONE)])  # Return value is of type `Quantity` during type-checking.
    Decimal('2')
    >>> isum([Amount(ONE), Amount(ONE)], Amount(ONE))  # Return value is of type `Amount` during type-checking.
    Decimal('3')
    >>> isum([Quantity(ONE), Quantity(ONE)], Quantity(ONE))  # Return value is of type `Quantity` during type-checking.
    Decimal('3')
    """
    return sum(xs, start or cast(DecimalLike, ZERO))


def sign(x: Numeric) -> int:
    """
    Defines a sign predicate for numeric values (ints, floats and Decimals).

    :param number: A number.
    :return: The sign of the given number.

    >>> sign(1)
    1
    >>> sign(0)
    0
    >>> sign(-0)
    0
    >>> sign(-1)
    -1
    >>> sign(Decimal("1"))
    1
    >>> sign(Decimal("0"))
    0
    >>> sign(-Decimal("0"))
    0
    >>> sign(Decimal("-1"))
    -1
    """
    return 1 if x > 0 else -1 if x < 0 else 0


def normalize(value: Decimal) -> Decimal:
    """
    Normalizes a given decimal value.

    :param value: The decimal value to be normalized.
    :return: The normalized decimal value.

    >>> normalize(Decimal("0.00"))
    Decimal('0')
    """
    return value.quantize(ONE) if value == value.to_integral() else value.normalize()


def weirdiv(dividend: Optional[Decimal], divisor: Optional[Decimal]) -> Decimal:
    """
    Provides a division for weirdos.

    Essentially, we are using this function at such call-sites where we don't have control over the incoming data, and
    lazy enough to bother doing it there.

    :param dividend: An optional dividend.
    :param divisor: An optional divisor.
    :return: A Decimal whether dividend and/or divisor are missing (0 in that case)

    >>> weirdiv(None, None)
    Decimal('0')
    >>> weirdiv(None, Decimal(0))
    Decimal('0')
    >>> weirdiv(None, Decimal(1))
    Decimal('0')
    >>> weirdiv(Decimal(0), None)
    Decimal('0')
    >>> weirdiv(Decimal(1), None) > 10 ** 10
    True
    >>> weirdiv(Decimal(9), Decimal(3))
    Decimal('3')
    """
    ## Check the dividend:
    if dividend is None or dividend.is_zero():
        return ZERO

    ## Check the divisor:
    if divisor is None or divisor.is_zero():
        ## Wish I could return infinity!
        return Decimal(sys.float_info.max).copy_sign(dividend)

    ## Normal division:
    return dividend / divisor



"""
This module provides definitions and functionality applicable to the broader accounting scope.
"""

__all__ = ["Balance"]

import datetime
from dataclasses import dataclass

from pypara.commons.numbers import Quantity


@dataclass(frozen=True)
class Balance:
    """
    Provides a value object model for encoding dated balances.
    """

    #: Date of the balance.
    date: datetime.date

    #: Value of the balance.
    value: Quantity



import datetime
from decimal import Decimal

import pytest

from pypara.currencies import Currencies
from pypara.monetary import IncompatibleCurrencyError, Money, NoMoney, NoneMoney, Price, SomeMoney

## Define some currencies:
eur = Currencies["EUR"]
usd = Currencies["USD"]

## Defines some Decimal quantities:
zero = Decimal("0")
half = Decimal("0.5")
one = Decimal("1")
two = Decimal("2")

## Define some dates:
today = datetime.date.today()
yesterday = today - datetime.timedelta(days=1)


def test_implementation() -> None:
    ## Define instances:
    _money = Money()  # type: ignore
    smoney = SomeMoney(usd, one, today)
    nmoney = NoneMoney()

    ## Check structure:
    assert _money.__slots__ == ()
    assert smoney.__slots__ == ()
    assert nmoney.__slots__ == ()
    assert not hasattr(_money, "__dict__")
    assert not hasattr(smoney, "__dict__")
    assert not hasattr(nmoney, "__dict__")

    ## Check types
    assert isinstance(Money.NA, Money)
    assert isinstance(Money.NA, NoneMoney)
    assert not isinstance(Money.NA, SomeMoney)
    assert not isinstance(Money.NA, Price)

    assert isinstance(_money, Money)
    assert not isinstance(_money, SomeMoney)
    assert not isinstance(_money, NoneMoney)

    assert isinstance(smoney, Money)
    assert isinstance(smoney, SomeMoney)
    assert not isinstance(smoney, NoneMoney)

    assert isinstance(nmoney, Money)
    assert not isinstance(nmoney, SomeMoney)
    assert isinstance(nmoney, NoneMoney)


def test_of() -> None:
    assert Money.of(usd, one, None) == Money.NA
    assert Money.of(usd, None, today) == Money.NA
    assert Money.of(usd, one, None) == Money.NA
    assert Money.of(usd, one, today) == SomeMoney(usd, one, today)
    assert Money.of(usd, Decimal("0.055"), today) == Money.of(usd, Decimal("0.06"), today)
    assert Money.of(usd, Decimal("0.045"), today) == Money.of(usd, Decimal("0.04"), today)


def test_is_equal() -> None:
    ## Vanilla:
    assert Money.NA.is_equal(NoMoney)
    assert Money.NA.is_equal(NoneMoney())
    assert not Money.NA.is_equal(Money.of(usd, zero, today))
    assert Money.of(usd, zero, today).is_equal(Money.of(usd, zero, today))
    assert Money.of(usd, half, today).is_equal(Money.of(usd, half, today))
    assert not Money.of(usd, zero, today).is_equal(Money.of(eur, zero, today))
    assert not Money.of(usd, zero, today).is_equal(Money.of(usd, half, today))
    assert not Money.of(usd, zero, today).is_equal(Money.of(usd, zero, yesterday))

    ## With operator overload:
    assert Money.NA == NoneMoney()
    assert Money.NA != Money.of(usd, zero, today)
    assert Money.of(usd, zero, today) == Money.of(usd, zero, today)
    assert Money.of(usd, half, today) == Money.of(usd, half, today)
    assert Money.of(usd, zero, today) != Money.of(eur, zero, today)
    assert Money.of(usd, zero, today) != Money.of(usd, half, today)
    assert Money.of(usd, zero, today) != Money.of(usd, zero, yesterday)


def test_to_boolean() -> None:
    ## Vanilla:
    assert not Money.NA.as_boolean()
    assert not Money.of(usd, zero, today).as_boolean()
    assert Money.of(usd, half, today).as_boolean()
    assert Money.of(usd, -half, today).as_boolean()

    ## With semantic overload
    assert not bool(Money.NA)
    assert not Money.of(usd, zero, today)
    assert Money.of(usd, half, today)
    assert Money.of(usd, -half, today)


def test_to_float() -> None:
    ## Vanilla:
    with pytest.raises(TypeError):
        Money.NA.as_float()
    assert Money.of(usd, half, today).as_float() == 0.5
    assert type(Money.of(usd, half, today).as_float()) == float

    ## With overload:
    with pytest.raises(TypeError):
        float(Money.NA)
    assert float(Money.of(usd, half, today)) == 0.5
    assert type(float(Money.of(usd, half, today))) == float


def test_to_integer() -> None:
    ## Vanilla:
    with pytest.raises(TypeError):
        int(Money.NA)
    assert int(Money.of(usd, half, today)) == 0
    assert type(int(Money.of(usd, half, today))) == int

    ## With overload:
    with pytest.raises(TypeError):
        Money.NA.as_integer()
    assert Money.of(usd, half, today).as_integer() == 0
    assert type(Money.of(usd, half, today).as_integer()) == int


def test_abs() -> None:
    ## Vanilla:
    assert Money.NA.abs() == Money.NA
    assert Money.of(usd, zero, today).abs() == Money.of(usd, zero, today)
    assert Money.of(usd, -one, today).abs() == Money.of(usd, +one, today)
    assert Money.of(usd, +one, today).abs() == Money.of(usd, +one, today)

    ## With overload:
    assert abs(Money.NA) == Money.NA
    assert abs(Money.of(usd, zero, today)) == Money.of(usd, zero, today)
    assert abs(Money.of(usd, -one, today)) == Money.of(usd, +one, today)
    assert abs(Money.of(usd, +one, today)) == Money.of(usd, +one, today)


def test_negative() -> None:
    ## Vanilla:
    assert Money.NA.negative() == Money.NA
    assert Money.of(usd, zero, today).negative() == Money.of(usd, zero, today)
    assert Money.of(usd, -one, today).negative() == Money.of(usd, +one, today)
    assert Money.of(usd, +one, today).negative() == Money.of(usd, -one, today)

    ## With overload:
    assert -Money.NA == Money.NA
    assert -Money.of(usd, zero, today) == Money.of(usd, zero, today)
    assert -Money.of(usd, -one, today) == Money.of(usd, +one, today)
    assert -Money.of(usd, +one, today) == Money.of(usd, -one, today)


def test_positive() -> None:
    ## Vanilla:
    assert Money.NA.positive() == Money.NA
    assert Money.of(usd, zero, today).positive() == Money.of(usd, zero, today)
    assert Money.of(usd, -one, today).positive() == Money.of(usd, -one, today)
    assert Money.of(usd, +one, today).positive() == Money.of(usd, +one, today)

    ## With overload:
    assert +Money.NA == Money.NA
    assert +Money.of(usd, zero, today) == Money.of(usd, zero, today)
    assert +Money.of(usd, -one, today) == Money.of(usd, -one, today)
    assert +Money.of(usd, +one, today) == Money.of(usd, +one, today)


def test_round() -> None:
    ## Vanilla:
    assert Money.NA.round(2) == Money.NA
    assert Money.of(usd, zero, today).round(2) == Money.of(usd, zero, today)
    assert Money.of(usd, -one, today).round(2) == Money.of(usd, -one, today)
    assert Money.of(usd, +one, today).round(2) == Money.of(usd, +one, today)

    ## Quick tests:
    assert Money.of(usd, Decimal("1.555"), today).round(2) == Money.of(usd, Decimal("1.56"), today)
    assert Money.of(usd, Decimal("1.545"), today).round(2) == Money.of(usd, Decimal("1.54"), today)

    ## With overload:
    assert round(Money.NA, 2) == Money.NA
    assert round(Money.of(usd, zero, today), 2) == Money.of(usd, zero, today)
    assert round(Money.of(usd, -one, today), 2) == Money.of(usd, -one, today)
    assert round(Money.of(usd, +one, today), 2) == Money.of(usd, +one, today)
    assert round(Money.of(usd, Decimal("1.555"), today), 2) == Money.of(usd, Decimal("1.56"), today)
    assert round(Money.of(usd, Decimal("1.545"), today), 2) == Money.of(usd, Decimal("1.54"), today)

    ## Extras:
    assert round(Money.of(usd, Decimal("0.545"), today), 0) == Money.of(usd, Decimal("1"), today)
    assert round(Money.of(usd, Decimal("1.545"), today), 0) == Money.of(usd, Decimal("2"), today)
    assert round(Money.of(usd, Decimal("0.545"), today), 1) == Money.of(usd, Decimal("0.5"), today)
    assert round(Money.of(usd, Decimal("1.545"), today), 1) == Money.of(usd, Decimal("1.5"), today)
    assert round(Money.of(usd, Decimal("0.45"), today), 1) == Money.of(usd, Decimal("0.4"), today)
    assert round(Money.of(usd, Decimal("1.45"), today), 1) == Money.of(usd, Decimal("1.4"), today)

    ## TODO: Following two are not really what round function signature says. mypy can't detect it!
    assert round(Money.of(usd, Decimal("1.4"), today)) == Money.of(usd, Decimal("1"), today)
    assert round(Money.of(usd, Decimal("1.5"), today)) == Money.of(usd, Decimal("2"), today)


def test_addition() -> None:
    ## First use `Money.NA`s:
    assert Money.NA.add(Money.NA) == Money.NA
    assert Money.NA.add(Money.of(usd, zero, today)) == Money.of(usd, zero, today)
    assert Money.of(usd, zero, today).add(Money.NA) == Money.of(usd, zero, today)

    ## Vanilla addition:
    assert Money.of(usd, zero, today).add(Money.of(usd, zero, today)) == Money.of(usd, zero, today)
    assert Money.of(usd, zero, today).add(Money.of(usd, one, today)) == Money.of(usd, one, today)
    assert Money.of(usd, one, today).add(Money.of(usd, one, today)) == Money.of(usd, two, today)
    assert Money.of(usd, one, today).add(Money.of(usd, -one, today)) == Money.of(usd, zero, today)

    ## Carry dates forward:
    assert Money.of(usd, zero, today).add(Money.of(usd, one, yesterday)) == Money.of(usd, one, today)
    assert Money.of(usd, zero, yesterday).add(Money.of(usd, one, today)) == Money.of(usd, one, today)

    ## Incompatible currency errors:
    with pytest.raises(IncompatibleCurrencyError):
        Money.of(usd, zero, today).add(Money.of(eur, zero, today))

    ## Operator overload:
    assert Money.NA + Money.NA == Money.NA
    assert Money.NA + Money.of(usd, zero, today) == Money.of(usd, zero, today)
    assert Money.of(usd, zero, today) + Money.NA == Money.of(usd, zero, today)
    assert Money.of(usd, zero, today) + Money.of(usd, one, today) == Money.of(usd, one, today)


def test_scalar_addition() -> None:
    ## First use `Money.NA`s:
    assert Money.NA.scalar_add(1) == Money.NA

    ## Vanilla addition:
    assert Money.of(usd, zero, today).scalar_add(1) == Money.of(usd, one, today)
    assert Money.of(usd, zero, today).scalar_add(1.0) == Money.of(usd, one, today)
    assert Money.of(usd, zero, today).scalar_add(one) == Money.of(usd, one, today)
    assert Money.of(usd, zero, today).scalar_add(-1) == Money.of(usd, -one, today)

    ## Extras:
    assert Money.of(usd, zero, today).scalar_add(0.5) == Money.of(usd, half, today)
    assert Money.of(usd, zero, today).scalar_add(Decimal("0.05")) == Money.of(usd, Decimal("0.05"), today)
    assert Money.of(usd, zero, today).scalar_add(Decimal("0.005")) == Money.of(usd, Decimal("0"), today)
    assert Money.of(usd, zero, today).scalar_add(Decimal("0.015")) == Money.of(usd, Decimal("0.02"), today)


def test_subtraction() -> None:
    ## First use `Money.NA`s:
    assert Money.NA.subtract(Money.NA) == Money.NA
    assert Money.NA.subtract(Money.of(usd, zero, today)) == Money.of(usd, zero, today)
    assert Money.of(usd, zero, today).subtract(Money.NA) == Money.of(usd, zero, today)

    ## Vanilla subtraction:
    assert Money.of(usd, zero, today).subtract(Money.of(usd, zero, today)) == Money.of(usd, zero, today)
    assert Money.of(usd, zero, today).subtract(Money.of(usd, one, today)) == Money.of(usd, -one, today)
    assert Money.of(usd, one, today).subtract(Money.of(usd, one, today)) == Money.of(usd, zero, today)
    assert Money.of(usd, one, today).subtract(Money.of(usd, -one, today)) == Money.of(usd, two, today)

    ## Carry dates forward:
    assert Money.of(usd, zero, today).subtract(Money.of(usd, one, yesterday)) == Money.of(usd, -one, today)
    assert Money.of(usd, zero, yesterday).subtract(Money.of(usd, one, today)) == Money.of(usd, -one, today)

    ## Incompatible currency errors:
    with pytest.raises(IncompatibleCurrencyError):
        Money.of(usd, zero, today).subtract(Money.of(eur, zero, today))

    ## Operator overload:
    assert Money.of(usd, zero, today) - Money.of(usd, one, today) == Money.of(usd, -one, today)
    assert Money.NA - Money.NA == Money.NA
    assert Money.NA - Money.of(usd, zero, today) == Money.of(usd, zero, today)
    assert Money.of(usd, zero, today) - Money.NA == Money.of(usd, zero, today)


def test_scalar_subtraction() -> None:
    ## First use `Money.NA`s:
    assert Money.NA.scalar_subtract(1) == Money.NA

    ## Vanilla subtraction:
    assert Money.of(usd, zero, today).scalar_subtract(1) == Money.of(usd, -one, today)
    assert Money.of(usd, zero, today).scalar_subtract(1.0) == Money.of(usd, -one, today)
    assert Money.of(usd, zero, today).scalar_subtract(one) == Money.of(usd, -one, today)
    assert Money.of(usd, zero, today).scalar_subtract(-1) == Money.of(usd, one, today)

    ## Operator overload:
    assert Money.of(usd, zero, today).scalar_subtract(1) == Money.of(usd, -one, today)
    assert Money.of(usd, zero, today).scalar_subtract(-1) == Money.of(usd, one, today)

    ## Extras:
    assert Money.of(usd, zero, today).scalar_subtract(0.5) == Money.of(usd, -half, today)
    assert Money.of(usd, zero, today).scalar_subtract(Decimal("0.05")) == -Money.of(usd, Decimal("0.05"), today)
    assert Money.of(usd, zero, today).scalar_subtract(Decimal("0.005")) == -Money.of(usd, Decimal("0"), today)
    assert Money.of(usd, zero, today).scalar_subtract(Decimal("0.015")) == -Money.of(usd, Decimal("0.02"), today)


def test_scalar_multiplication() -> None:
    ## First use `Money.NA`s:
    assert Money.NA.multiply(1) == Money.NA

    ## Vanilla subtraction:
    assert Money.of(usd, one, today).multiply(1) == Money.of(usd, one, today)
    assert Money.of(usd, one, today).multiply(2) == Money.of(usd, two, today)
    assert Money.of(usd, -one, today).multiply(1) == Money.of(usd, -one, today)
    assert Money.of(usd, -one, today).multiply(2) == Money.of(usd, -two, today)

    ## Other types:
    assert Money.of(usd, one, today).multiply(1) == Money.of(usd, one, today)
    assert Money.of(usd, one, today).multiply(1.0) == Money.of(usd, one, today)
    assert Money.of(usd, one, today).multiply(one) == Money.of(usd, one, today)

    ## Operator overload:
    assert Money.NA * 1 == Money.NA
    assert Money.of(usd, one, today) * 1 == Money.of(usd, one, today)
    assert Money.of(usd, one, today) * 2 == Money.of(usd, two, today)
    assert Money.of(usd, -one, today) * 1 == Money.of(usd, -one, today)
    assert Money.of(usd, -one, today) * 2 == Money.of(usd, -two, today)

    ## Extras
    assert Money.of(usd, one, today).multiply(Decimal("0.050")) == Money.of(usd, Decimal("0.05"), today)
    assert Money.of(usd, one, today).multiply(Decimal("0.005")) == Money.of(usd, Decimal("0.00"), today)
    assert Money.of(usd, one, today).multiply(Decimal("0.015")) == Money.of(usd, Decimal("0.02"), today)


def test_division() -> None:
    ## First use `Money.NA`s:
    assert Money.NA.divide(1) == Money.NA

    ## Vanilla subtraction:
    assert Money.of(usd, one, today).divide(1) == Money.of(usd, one, today)
    assert Money.of(usd, one, today).divide(2) == Money.of(usd, half, today)
    assert Money.of(usd, -one, today).divide(1) == Money.of(usd, -one, today)
    assert Money.of(usd, -one, today).divide(2) == Money.of(usd, -half, today)

    ## Various divisor types:
    assert Money.of(usd, one, today).divide(2) == Money.of(usd, half, today)
    assert Money.of(usd, one, today).divide(2.0) == Money.of(usd, half, today)
    assert Money.of(usd, one, today).divide(two) == Money.of(usd, half, today)

    ## Division by zero:
    assert Money.of(usd, one, today).divide(0) == Money.NA
    assert Money.of(usd, one, today).divide(zero) == Money.NA
    assert Money.of(usd, one, today).divide(0.0) == Money.NA

    ## Operator overload:
    assert Money.NA / 1 == Money.NA
    assert Money.of(usd, one, today) / 1 == Money.of(usd, one, today)
    assert Money.of(usd, one, today) / 2 == Money.of(usd, half, today)
    assert Money.of(usd, -one, today) / 1 == Money.of(usd, -one, today)
    assert Money.of(usd, -one, today) / 2 == Money.of(usd, -half, today)
    assert Money.of(usd, -one, today) / 0 == Money.NA

    ## Extras
    assert Money.of(usd, one, today).divide(Decimal("10")) == Money.of(usd, Decimal("0.10"), today)
    assert Money.of(usd, one, today).divide(Decimal("50")) == Money.of(usd, Decimal("0.02"), today)
    assert Money.of(usd, one, today).divide(Decimal("100")) == Money.of(usd, Decimal("0.01"), today)
    assert Money.of(usd, one, today).divide(Decimal("1000")) == Money.of(usd, Decimal("0.00"), today)


def test_floor_division() -> None:
    ## First use `Money.NA`s:
    assert Money.NA.floor_divide(1) == Money.NA

    ## Vanilla subtraction:
    assert Money.of(usd, one, today).floor_divide(1) == Money.of(usd, one, today)
    assert Money.of(usd, one, today).floor_divide(2) == Money.of(usd, zero, today)
    assert Money.of(usd, -one, today).floor_divide(1) == Money.of(usd, -one, today)
    assert Money.of(usd, -one, today).floor_divide(2) == Money.of(usd, zero, today)

    ## Various divisor types:
    assert Money.of(usd, one, today).floor_divide(2) == Money.of(usd, zero, today)
    assert Money.of(usd, one, today).floor_divide(2.0) == Money.of(usd, zero, today)
    assert Money.of(usd, one, today).floor_divide(two) == Money.of(usd, zero, today)

    ## Division by zero:
    assert Money.of(usd, one, today).floor_divide(0) == Money.NA
    assert Money.of(usd, one, today).floor_divide(zero) == Money.NA
    assert Money.of(usd, one, today).floor_divide(0.0) == Money.NA

    ## Operator overload:
    assert Money.NA / 1 == Money.NA
    assert Money.of(usd, one, today) // 1 == Money.of(usd, one, today)
    assert Money.of(usd, one, today) // 2 == Money.of(usd, zero, today)
    assert Money.of(usd, -one, today) // 1 == Money.of(usd, -one, today)
    assert Money.of(usd, -one, today) // 2 == Money.of(usd, zero, today)
    assert Money.of(usd, -one, today) // 0 == Money.NA

    ## Extras
    assert Money.of(usd, Decimal("10"), today).floor_divide(Decimal("10")) == Money.of(usd, Decimal("1.00"), today)
    assert Money.of(usd, Decimal("10"), today).floor_divide(Decimal("11")) == Money.of(usd, Decimal("0.00"), today)


def test_comparisons() -> None:
    ## First use `Money.NA`s:
    assert not (Money.NA < Money.NA)
    assert Money.NA <= Money.NA
    assert not (Money.NA > Money.NA)
    assert Money.NA >= Money.NA

    ## Try mixed:
    assert Money.NA < Money.of(usd, -one, today)
    assert Money.NA <= Money.of(usd, -one, today)
    assert not (Money.NA > Money.of(usd, -one, today))
    assert not (Money.NA >= Money.of(usd, -one, today))

    ## ... and:
    assert not (Money.of(usd, -one, today) < Money.NA)
    assert not (Money.of(usd, -one, today) <= Money.NA)
    assert Money.of(usd, -one, today) > Money.NA
    assert Money.of(usd, -one, today) >= Money.NA

    ## With defined values:
    assert not (Money.of(usd, zero, today) < Money.of(usd, zero, today))
    assert Money.of(usd, zero, today) <= Money.of(usd, zero, today)
    assert not (Money.of(usd, zero, today) > Money.of(usd, zero, today))
    assert Money.of(usd, zero, today) >= Money.of(usd, zero, today)

    ## ... and:
    assert Money.of(usd, zero, today) < Money.of(usd, one, today)
    assert Money.of(usd, zero, today) <= Money.of(usd, one, today)
    assert not (Money.of(usd, zero, today) > Money.of(usd, one, today))
    assert not (Money.of(usd, zero, today) >= Money.of(usd, one, today))

    ## ... and:
    assert not (Money.of(usd, one, today) < Money.of(usd, zero, today))
    assert not (Money.of(usd, one, today) <= Money.of(usd, zero, today))
    assert Money.of(usd, one, today) > Money.of(usd, zero, today)
    assert Money.of(usd, one, today) >= Money.of(usd, zero, today)


def test_with() -> None:
    ## First use `Money.NA`s:
    assert Money.NA.with_ccy(usd) == Money.NA
    assert Money.NA.with_qty(one) == Money.NA
    assert Money.NA.with_dov(today) == Money.NA

    ## Now with some:
    assert Money.of(usd, zero, today).with_ccy(eur) == Money.of(eur, zero, today)
    assert Money.of(usd, zero, today).with_qty(one) == Money.of(usd, one, today)
    assert Money.of(usd, zero, today).with_dov(yesterday) == Money.of(usd, zero, yesterday)

    ## Extras:
    assert Money.of(usd, zero, today).with_qty(Decimal("0.005")) == Money.of(usd, zero, today)
    assert Money.of(usd, zero, today).with_qty(Decimal("0.054")) == Money.of(usd, Decimal("0.05"), today)
